# eSCAPyDL - a Deep Learning Side-Channel Analysis Python Framework
# Copyright (C) 2023  Weissbart Léo

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os,sys
import argparse
import time
import signal
import shlex
import subprocess
import webbrowser
import gi
import json
import logging
import re
logging.basicConfig(format='%(levelname)s:%(message)s')
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
import fcntl

sys.path.insert(0, os.path.dirname(os.path.realpath(__file__)))
from datasets.to_json import to_json

gi.require_version("Gtk", "4.0")
gi.require_version('GtkSource', '5')
gi.require_version('Adw', '1')
gi.require_version('Gdk', '4.0')
from gi.repository import GLib, Gtk, GtkSource, Adw, Gio, Gdk

class gtkFileMonitorWindow(Gtk.ScrolledWindow):
    def __init__(self, file_path):
        super().__init__()
        self.set_vexpand(True)
        self.file_path = file_path
        #open the file
        file = Gio.File.new_for_path(file_path)
        self.file_monitor = file.monitor_file(Gio.FileMonitorFlags.NONE, None)
        self.file_monitor.connect("changed", self.file_changed)
        
        # Create a text view to show the file content
        self.textview = GtkSource.View()
        self.set_child(self.textview)
        self.textview.buffer = GtkSource.Buffer()
        self.textview.set_editable(False)
        self.textview.set_buffer(self.textview.buffer)

        style_provider = Gtk.CssProvider()
        style_provider.load_from_resource("/org/gnome/escapydl/themes/dark-theme.css")
        self.textview.get_style_context().add_provider(style_provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)

        self.log_linter()
    
    def log_linter(self):
        # Create a simple linter for log files
        # Create tags for various elements
        self.textview.numeric = self.textview.buffer.create_tag("numeric", foreground="#4b6cd0")
        self.textview.logging = self.textview.buffer.create_tag("logging", foreground="#50fa7b")
        self.textview.filepath = self.textview.buffer.create_tag("filepath", foreground="#d97c1a")
        # Connect the buffer's "changed" signal to update the linter
        self.textview.buffer.connect("changed", self.update_linter)
    
    def update_linter(self, buffer):
        # Clear all existing tags
        buffer.remove_all_tags(buffer.get_start_iter(), buffer.get_end_iter())
        # Get the current text from the buffer
        text = buffer.get_text(buffer.get_start_iter(), buffer.get_end_iter(), False)
        # Apply tags for file path
        for match in re.finditer(r"\"[^\s]+\"", text):
            start, end = match.span(0)
            buffer.apply_tag(self.textview.filepath, buffer.get_iter_at_offset(start), buffer.get_iter_at_offset(end))
        # Apply tags for numeric
        for match in re.finditer(r"\b\d+\.*\d*\b", text):
            start, end = match.span(0)
            buffer.apply_tag(self.textview.numeric, buffer.get_iter_at_offset(start), buffer.get_iter_at_offset(end))
        # Apply tags for logging Does not contain a tag but only timestamp
        for match in re.finditer(r"\d{2}\/\d{2}\/\d{4}\s\d{2}:\d{2}:\d{2}", text):
            start, end = match.span(0)
            buffer.apply_tag(self.textview.logging, buffer.get_iter_at_offset(start), buffer.get_iter_at_offset(end))

    def file_changed(self, file_monitor, file, other_file, event_type):
        # This function is called when the file is modified
        # Get the difference between the new text and the old text and write it to the textview
        new_text = file.load_contents(None)[1].decode()
        old_text = self.textview.buffer.get_text(self.textview.buffer.get_start_iter(), self.textview.buffer.get_end_iter(), False)
        new_text = new_text[len(old_text):]
        if new_text != "":
            self.textview.buffer.insert(self.textview.buffer.get_end_iter(), new_text)
            end_mark = self.textview.buffer.create_mark(None, self.textview.buffer.get_end_iter(), True)
            self.textview.scroll_to_mark(end_mark, 0, False, 0, 0)

class gtkPipeMonitorWindow(Gtk.ScrolledWindow):
    def __init__(self):
        super().__init__()
        self.sub_proc = None
        self.set_vexpand(True)

        # Set a sticky flag to keep the scroll at the bottom
        self.sticky = True
        self.prev_value_vposition = 0
        self.vadjustment = self.get_vadjustment()
        self.vadjustment.connect("value-changed", self.on_scroll_child)
        
        # Create a text view to show the file content
        self.textview = GtkSource.View()
        self.set_child(self.textview)
        self.textview.buffer = GtkSource.Buffer()
        self.textview.set_editable(False)
        self.textview.set_buffer(self.textview.buffer)
        self.textview.textiter = self.textview.buffer.get_end_iter()

        self.is_progress = False

        style_provider = Gtk.CssProvider()
        style_provider.load_from_resource("/org/gnome/escapydl/themes/dark-theme.css")
        self.textview.get_style_context().add_provider(style_provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)

        self.log_linter()


    def on_scroll_child(self, adjustment):
        # If the user scrolls up, disable sticky, enable if the user scrolls down to bottom
        current_value = adjustment.get_value()
        if current_value < self.prev_value_vposition:
            self.sticky = False
        elif current_value > self.prev_value_vposition and adjustment.get_value() == adjustment.get_upper() - adjustment.get_page_size():
            self.sticky = True
        self.prev_value_vposition = current_value

    def non_block_read(self, output):
        ''' even in a thread, a normal read with block until the buffer is full '''
        fd = output.fileno()
        fl = fcntl.fcntl(fd, fcntl.F_GETFL)
        fcntl.fcntl(fd, fcntl.F_SETFL, fl | os.O_NONBLOCK)
        try:
            return output.read().decode("utf-8")
        except:
            return ''

    def update_terminal(self):
        self.textview.textiter = self.textview.buffer.get_end_iter()
        new_text = self.non_block_read(self.sub_proc.stdout)
        # logger.debug(repr(new_text))
        if new_text.endswith('\\r\n'): # nasty trick to show progressbar
            new_text = new_text.split('\n')[-2][:-2]+'\n'
            self.textview.textiter.backward_line()
            self.textview.buffer.delete(self.textview.textiter,self.textview.buffer.get_end_iter())
        self.textview.buffer.insert(self.textview.textiter, new_text)
        self.textview.textiter = self.textview.buffer.get_end_iter()
        if self.sticky:
            self.vadjustment.set_value(self.vadjustment.get_upper() - self.vadjustment.get_page_size())
        return self.sub_proc.poll() is None
    
    def attach_sub_proc(self, sub_proc):
        self.sub_proc = sub_proc
        GLib.timeout_add(100, self.update_terminal)
        logger.debug("Sub process attached to monitor window")

    def log_linter(self):
        # Create a simple linter for log files
        # Create tags for various elements
        self.textview.numeric = self.textview.buffer.create_tag("numeric", foreground="#4b6cd0")
        self.textview.logging = self.textview.buffer.create_tag("logging", foreground="#50fa7b")
        self.textview.filepath = self.textview.buffer.create_tag("filepath", foreground="#d97c1a")
        # Connect the buffer's "changed" signal to update the linter
        self.textview.buffer.connect("changed", self.update_linter)
    
    def update_linter(self, buffer):
        # Clear all existing tags
        buffer.remove_all_tags(buffer.get_start_iter(), buffer.get_end_iter())
        # Get the current text from the buffer
        text = buffer.get_text(buffer.get_start_iter(), buffer.get_end_iter(), False)
        # Apply tags for file path
        for match in re.finditer(r"\"[^\s]+\"", text):
            start, end = match.span(0)
            buffer.apply_tag(self.textview.filepath, buffer.get_iter_at_offset(start), buffer.get_iter_at_offset(end))
        # Apply tags for numeric
        for match in re.finditer(r"\b\d+\.*\d*\b", text):
            start, end = match.span(0)
            buffer.apply_tag(self.textview.numeric, buffer.get_iter_at_offset(start), buffer.get_iter_at_offset(end))
        # Apply tags for logging Does not contain a tag but only timestamp
        for match in re.finditer(r"\d{2}\/\d{2}\/\d{4}\s\d{2}:\d{2}:\d{2}", text):
            start, end = match.span(0)
            buffer.apply_tag(self.textview.logging, buffer.get_iter_at_offset(start), buffer.get_iter_at_offset(end))

class gtkGridbox(Gtk.Grid):
    def __init__(self):
        super().__init__()
        self.set_margin_start(5)
        self.set_margin_end(5)
        self.set_column_spacing(6)
        self.set_row_spacing(6)

class gtkAddRemoveGridbox(gtkGridbox):
    def __init__(self, add_button_function, remove_button_function):
        super().__init__()
        addlayerbutton = Gtk.Button.new_with_label("+")
        addlayerbutton.connect("clicked", add_button_function)
        self.attach(addlayerbutton, 0, 0, 1, 1)
        rmlayerbutton = Gtk.Button.new_with_label("-")
        rmlayerbutton.connect("clicked", remove_button_function)
        self.attach(rmlayerbutton, 1, 0, 1, 1)

class gtkComboText_with_items(Gtk.ComboBoxText):
    def __init__(self, items, set_active=0):
        super().__init__()
        for item in items:
            self.append_text(item)
        if set_active != None:
            self.set_active(set_active)

class gtkCodebox(Gtk.ScrolledWindow):
    def __init__(self):
        super().__init__()
        self.set_hexpand(True)
        self.set_vexpand(True)
        self.set_size_request(-1, 150)
        self.set_margin_start(5)
        self.set_margin_end(5)
        
        self.textview = GtkSource.View()
        self.set_child(self.textview)
        self.textview.set_auto_indent(True)
        self.textview.set_indent_width(4)
        self.textview.set_accepts_tab(True)
        self.textview.set_wrap_mode(Gtk.WrapMode.WORD)

        style_provider = Gtk.CssProvider()
        style_provider.load_from_resource("/org/gnome/escapydl/themes/dark-theme.css")
        self.textview.get_style_context().add_provider(style_provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)

        # Set up the source buffer for the text view
        self.textview.buffer = GtkSource.Buffer()
        self.textview.set_buffer(self.textview.buffer)
        self.python_linter()
    
    def python_linter(self):
        # Create a simple linter for Python code

        # Create a language manager and set the language to Python
        lang_manager = GtkSource.LanguageManager()
        lang = lang_manager.get_language("python")
        self.textview.buffer.set_language(lang)

        # Create tags for various elements
        self.textview.function_tag = self.textview.buffer.create_tag("function", foreground="#E0BA30")#yellow
        self.textview.variable_tag = self.textview.buffer.create_tag("variable", foreground="#2F7AE3")#blue
        self.textview.package_tag = self.textview.buffer.create_tag("type", foreground="#2F7AE3")#blue
        self.textview.subpackage_tag = self.textview.buffer.create_tag("subpackage", foreground="#2F7AE3")#blue
        self.textview.subsubpackage_tag = self.textview.buffer.create_tag("subsubpackage", foreground="#2F7AE3")#blue
        self.textview.string_tag = self.textview.buffer.create_tag("string", foreground="#d97c1a")#orange
        # Connect the buffer's "changed" signal to update the linter
        self.textview.buffer.connect("changed", self.update_linter)
    
    def update_linter(self, buffer):
        # Clear all existing tags
        buffer.remove_all_tags(buffer.get_start_iter(), buffer.get_end_iter())
        # Get the current text from the buffer
        text = buffer.get_text(buffer.get_start_iter(), buffer.get_end_iter(), False)
        # Apply tags for functions
        for match in re.finditer(r"def\s+([a-zA-Z_]\w*)\((.*)\)\:", text):
            start, end = match.span(1)
            buffer.apply_tag(self.textview.function_tag, buffer.get_iter_at_offset(start), buffer.get_iter_at_offset(end))
            arguments_text = match.group(2)
            arguments_start = match.start(2)
            for arg in re.finditer(r"\b\w+\b", arguments_text): 
                start, end = arg.span(0)
                buffer.apply_tag(self.textview.variable_tag, buffer.get_iter_at_offset(arguments_start+start), buffer.get_iter_at_offset(arguments_start+end))
        # Apply tags for package names (up to three sub elements)
        for match in re.finditer(r"(\w+)\.(\w+)\.?(\w+)?", text):
            start, end = match.span(1)
            buffer.apply_tag(self.textview.package_tag, buffer.get_iter_at_offset(start), buffer.get_iter_at_offset(end))
            if match.span(2) != (0,0):
                start, end = match.span(2)
                buffer.apply_tag(self.textview.subpackage_tag, buffer.get_iter_at_offset(start), buffer.get_iter_at_offset(end))
            if match.span(3) != (-1,-1):
                start, end = match.span(3)
                buffer.apply_tag(self.textview.subsubpackage_tag, buffer.get_iter_at_offset(start), buffer.get_iter_at_offset(end))
        # Apply tags for any string
        for match in re.finditer(r"[\'\"]\w+?[\'\"]", text):
            start, end = match.span(0)
            buffer.apply_tag(self.textview.string_tag, buffer.get_iter_at_offset(start), buffer.get_iter_at_offset(end))

class gtkVerticalBox(Gtk.Box):
    def __init__(self):
        super().__init__(orientation=Gtk.Orientation.VERTICAL, spacing=6)
        self.set_margin_start(5)
        self.set_margin_end(5)
        self.set_margin_top(5)
        self.set_margin_bottom(5)

class gtkResizableEntry(Gtk.Entry):
    def __init__(self,label=None):
        super().__init__()
        if label != None:
            self.set_buffer(Gtk.EntryBuffer.new(label,-1))
        self.connect("changed", self.on_entry_changed)
    
    def on_entry_changed(self,widget):
        self.set_width_chars(len(self.get_text()))

class gtkHorizontalBox(Gtk.Box):
    def __init__(self):
        super().__init__(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
        self.set_margin_start(5)
        self.set_margin_end(5)
        self.set_margin_top(5)
        self.set_margin_bottom(5)

class gtkLabelLeftAlign(Gtk.Label):
    def __init__(self, label):
        super().__init__(label=label)
        self.set_halign(Gtk.Align.START)

class gtkLabelWithTooltip(Gtk.Grid):
    #Makes a grid containing a label and an interogation mark icon next to it with a tooltip
    def __init__(self, label, tooltip):
        super().__init__()
        self.set_column_spacing(6)
        self.set_row_spacing(6)
        self.set_halign(Gtk.Align.START)
        self.set_valign(Gtk.Align.CENTER)
        self.label = Gtk.Label(label=label)
        self.label.set_halign(Gtk.Align.START)
        self.label.set_valign(Gtk.Align.CENTER)
        self.attach(self.label, 0, 0, 1, 1)
        self.icon = Gtk.Image.new_from_icon_name("info-symbolic")
        self.icon.set_halign(Gtk.Align.START)
        self.icon.set_valign(Gtk.Align.CENTER)
        self.attach(self.icon, 1, 0, 1, 1)
        self.icon.set_tooltip_text(tooltip)

    def get_buffer(self):
        return self.label.get_buffer()

class gtkButtonFileSelection(Gtk.Button):
    def __init__(self, entry, file_path=None):
        super().__init__()
        self.set_halign(Gtk.Align.START)
        self.set_valign(Gtk.Align.CENTER)
        self.set_icon_name("document-open-symbolic")
        self.connect("clicked", self.on_file_clicked)
        self.file_path = file_path
        self.entry = entry
        self.style_provider = Gtk.CssProvider()
        self.style_provider.load_from_resource("/org/gnome/escapydl/themes/dark-theme.css")

    def set_file_path(self, file_path):
        self.file_path = file_path
    
    def on_file_clicked(self, widget):
        self.window = Gtk.Window(title="Select file")
        self.window.set_icon_name("escapydl-icon")
        self.window.set_default_size(900, 600)
        content = gtkVerticalBox()
        self.dialog = Gtk.FileChooserWidget.new(0)
        self.dialog.set_current_folder(Gio.File.parse_name(os.getcwd()))
        self.dialog.set_vexpand(True)
        content.append(self.dialog)
        buttons = Gtk.HeaderBar.new()
        buttons.set_show_title_buttons(False)
        # buttons.append(Gtk.Box(hexpand=True))
        select_button = Gtk.Button.new_with_label("Select")
        select_button.get_style_context().add_provider(self.style_provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)
        select_button.set_name("button_green")
        select_button.connect("clicked", self.select_button_clicked)
        buttons.pack_end(select_button)
        cancel_button = Gtk.Button.new_with_label("Cancel")
        cancel_button.connect("clicked", lambda x: self.window.close())
        buttons.pack_start(cancel_button)
        # content.append(buttons)
        self.window.set_titlebar(buttons)
        self.window.set_child(content)
        self.window.present()

    def select_button_clicked(self, widget):
        self.file_path = self.dialog.get_file().get_path()
        self.entry.set_buffer(Gtk.EntryBuffer(text=self.file_path))
        self.entry.on_entry_changed(self.entry)
        logger.debug("File selected: " + self.file_path)

        self.window.close()

class eSCAPyDLgui(Gtk.Application):
    def __init__(self, timestamp=time.strftime("%Y%m%d_%H%M%S",time.localtime()), log_filename='', verbose=0):
        super().__init__(application_id="org.gnome.escapydl")
        self.timestamp = timestamp
        self.log_filename = log_filename
        self.verbose = verbose
        GLib.set_application_name("eSCAPyDL")
        self.style_manager = Adw.StyleManager().get_default()
        self.style_manager.set_color_scheme(Adw.ColorScheme.PREFER_DARK)

        # Create an icon theme with local icons for better transferability
        resource_filename = os.path.join(os.path.dirname(__file__), "resources/resource.gresource")
        resource = Gio.Resource.load(resource_filename)
        Gio.resources_register(resource)
        self.icon_theme = Gtk.IconTheme.get_for_display(Gdk.Display.get_default())
        self.icon_theme.add_resource_path("/org/gnome/escapydl/icons")
        
        self.window = None
        self.outerbox = Gtk.Box(spacing=6)
        self.dataset = None
        self.study_name = ""

        self.template_models = ["Specific Model", "MLP", "CNN", "ASCAD_mlp", "ASCAD_cnn"]
        self.json_data = {}
        self.parameter_fields = ['dataset_name', 'dataset_size', 'training_size', 'val_size', 'subkey', 'num_key_hypotheses', 'n_samples', 'shuffle']
        self.parameter_fields_wval = ['dataset_name', 'validation_name', 'dataset_size', 'training_size', 'val_size', 'subkey', 'num_key_hypotheses', 'n_samples', 'shuffle']
        self.callbacks_fileds = ['ModelCheckpoint', 'Tensorboard', 'GuessingEntropy']
        self.functions_fields = ['get_attack_parameters', 'get_known_key', 'intermediate_value']
        self.script_action = 'train'
        self.modeltype = None
        self.mlp_parameters = ['num_epochs', 'bs', 'loss_function', 'lr', 'activation_function', 'pooling', 'optimizer', 'weight_init']
        self.cnn_parameters = ['num_epochs', 'bs', 'loss_function', 'lr', 'activation_function', 'pooling', 'optimizer', 'weight_init']
        self.n_conv_block = 0
        self.optuna_database_name = "sqlite:///cnn_sca.db"

    def do_activate(self):
        self.window = Gtk.ApplicationWindow(application=self, title="eSCAPyDL",default_width=800, default_height=600)
        self.window.set_icon_name("escapydl-icon")
        self.window.connect("close-request", self.on_button_close_clicked)
        self.to_first_window(self)

    def add_navigation_bar(self, prev_action, next_action,fill_gap=True):
        if fill_gap:
            #Fill with empty box
            dbox = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, spacing=6)
            dbox.set_vexpand(True)
            self.outerbox.append(dbox)
        # Add previous/next button on the bottom of the window
        bbox = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, spacing=6)
        pbutton = Gtk.Button.new_with_label("Previous")
        pbutton.set_halign(Gtk.Align.START)
        pbutton.connect("clicked", prev_action)
        bbox.append(pbutton)
        fillbox = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, spacing=6)
        fillbox.set_hexpand(True)
        bbox.append(fillbox)
        nbutton = Gtk.Button.new_with_label("Next")
        nbutton.set_halign(Gtk.Align.END)
        nbutton.connect("clicked", next_action)
        bbox.append(nbutton)
        # bbox.set_baseline_position(Gtk.BaselinePosition.BOTTOM)
        self.outerbox.append(bbox)

    def on_button_clicked(self, widget):
        logger.info("Hello World")

    def on_button_validationdataset_clicked(self, button):
        if button.get_active():
            self.use_validation_dataset = False
            #Remove the line below
            self.paramgridbox.remove_row(button.row_index+1)
        else:
            self.use_validation_dataset = True
            #Add a newline in paramGridBox and create a field for validation dataset name
            self.paramgridbox.insert_row(button.row_index+1)
            self.paramgridbox.attach(gtkLabelLeftAlign(label="validation_name: "), 0, button.row_index+1, 1, 1)
            entryval = gtkResizableEntry(self.json_data['parameters']['validation_name'] if 'validation_name' in self.json_data['parameters'].keys() else "")
            self.paramgridbox.attach(entryval, 1, button.row_index+1, 1, 1)
            self.paramgridbox.attach(gtkButtonFileSelection(entry=entryval), 2, button.row_index+1, 1, 1)

    def on_button_close_clicked(self, widget):
        logger.debug("Closing window")
        try:
            logger.debug("Killing child process {}".format(self.child_process.pid))
            self.child_process.send_signal(signal.SIGINT)
        except AttributeError as e:
            logger.debug("No child process started")
        try:
            logger.debug("Killing tensorboard child process {}".format(self.tb_child_process.pid))
            self.tb_child_process.send_signal(signal.SIGINT)
        except AttributeError as e:
            logger.debug("No tensorboard child process started")
        try:
            logger.debug("Killing optuna child process {}".format(self.optuna_child_process.pid))
            self.optuna_child_process.send_signal(signal.SIGINT)
        except AttributeError as e:
            logger.debug("No optuna child process started")
        self.window.close()

    def on_button_tensorboarb_clicked(self, widget):
        command = "tensorboard --logdir=runs"
        logger.debug("Starting tensorboard")
        try:
            if not hasattr(self, "tb_child_process") or self.tb_child_process.poll() is not None:
                self.tb_child_process = subprocess.Popen(shlex.split(command))
                logger.debug("Tensorboard child process created, id:{}".format(self.tb_child_process.pid))
            else:
                logger.debug("Tensorboard child process already running, id:{}".format(self.tb_child_process.pid))
            time.sleep(1)
            webbrowser.open_new("http://localhost:6006/?runFilter={}".format(self.study_name))
        except Exception as e:
            logger.warning(e)

    def on_button_optuna_dashboard_clicked(self, widget):
        if os.path.isfile(os.path.join(os.getcwd(),self.optuna_database_name[10:])):
            command = "optuna-dashboard {} --port 2021".format(self.optuna_database_name)
            logger.debug("Starting optuna dashboard")
            try:
                if not hasattr(self, "optuna_child_process") or self.optuna_child_process.poll() is not None:
                    self.optuna_child_process = subprocess.Popen(shlex.split(command))
                    logger.debug("Optuna dashboard child process created, id:{}".format(self.optuna_child_process.pid))
                else:
                    logger.debug("Optuna dashboard child process already running, id:{}".format(self.optuna_child_process.pid))
                time.sleep(1)
                webbrowser.open_new("http://localhost:2021")
            except Exception as e:
                logger.warning(e)
        else:
            logger.warning("Optuna database not found: {}".format(os.path.join(os.getcwd(),self.optuna_database_name[10:])))

    def on_button_run_script_clicked(self, widget):
        logger.debug("Run script")
        self.output_window.textview.buffer.set_text("")
        self.timestamp=time.strftime("%Y%m%d_%H%M%S",time.localtime())#refresh timestamp
        self.study_name = "{}/{}/{}".format(self.dataset.lower(),self.modeltype,self.timestamp)
        main_py_filename = os.path.join(os.path.dirname(__file__), "main.py")
        command = sys.executable + " {} --{} -d {} -m {} -j {} -s{} -a {} -l '{}'".format(main_py_filename, self.script_action, str.lower(self.dataset), self.modeltype, self.json_filename, self.study_name, self.timestamp, self.log_filename)
        if self.verbose > 0:
            command += " -v"
        logger.debug("Command: {}".format(command))
        try:
            self.child_process = subprocess.Popen(shlex.split(command), stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            logger.debug("Child process created, id:{}".format(self.child_process.pid))
            self.output_window.attach_sub_proc(self.child_process)
            widget.set_sensitive(False)
        except Exception as e:
            logger.warning(e)

    def on_button_stop_script_clicked(self, widget):
        logger.debug("Stop script")
        if hasattr(self, "child_process") and isinstance(self.child_process, subprocess.Popen):
            logger.debug("Child process killed, id:{}".format(self.child_process.pid))
            self.child_process.send_signal(signal.SIGINT)
            if self.child_process.poll() is not None:
                self.child_process = None
            widget.get_parent().get_first_child().set_sensitive(True)#Set run button sensitive again
        try:
            logger.debug("Killing tensorboard child process {}".format(self.tb_child_process.pid))
            self.tb_child_process.send_signal(signal.SIGINT)
        except AttributeError as e:
            logger.debug("No tensorboard child process started")
        try:
            logger.debug("Killing optuna-dashboard child process {}".format(self.optuna_child_process.pid))
            self.optuna_child_process.send_signal(signal.SIGINT)
        except AttributeError as e:
            logger.debug("No tensorboard child process started")

    def on_button_trainopt_clicked(self, button):
        if button.get_label().lower() == "train":
            self.script_action = 'train'
            logger.debug("Train button is active")
        elif button.get_label().lower() == "optimize":
            self.script_action = 'optimize'
            logger.debug("Optimize button is active")
        self.refresh_listbox_content()
    
    def on_button_addlayer_clicked(self, widget):
        #insert a new row in the gridbox
        self.mlp_layers_box.insert_row(self.json_data['model']['n_layers'])
        self.mlp_layers_box.attach(Gtk.Label(label="Layer {}".format(self.json_data['model']['n_layers']+1)), 0, self.json_data['model']['n_layers'], 1, 1)
        previous_layer_text = self.mlp_layers_box.get_child_at(1,self.json_data['model']['n_layers']-1).get_text()
        self.mlp_layers_box.attach(Gtk.Entry.new_with_buffer(Gtk.EntryBuffer.new(previous_layer_text,-1)), 1, self.json_data['model']['n_layers'], 1, 1)
        self.json_data['model']['n_layers'] += 1

    def on_button_removelayer_clicked(self, widget):
        if self.json_data['model']['n_layers'] > 1:
            self.mlp_layers_box.remove_row(self.json_data['model']['n_layers']-1)
            self.json_data['model']['n_layers'] -= 1

    def on_button_addconv_clicked(self, widget):
        #insert a new row in the gridbox
        self.n_conv_block += 1
        self.cnn_block_box.insert_row(self.n_conv_block)
        self.cnn_block_box.attach(Gtk.Label(label="Conv {}".format(self.n_conv_block)), 0, self.n_conv_block, 1, 1)
        previous_layer_text1 = self.cnn_block_box.get_child_at(1,self.n_conv_block-1).get_text()
        self.cnn_block_box.attach(Gtk.Entry.new_with_buffer(Gtk.EntryBuffer.new(previous_layer_text1,-1)), 1, self.n_conv_block, 1, 1)
        previous_layer_text2 = self.cnn_block_box.get_child_at(2,self.n_conv_block-1).get_text()
        self.cnn_block_box.attach(Gtk.Entry.new_with_buffer(Gtk.EntryBuffer.new(previous_layer_text2,-1)), 2, self.n_conv_block, 1, 1)

    def on_button_removeconv_clicked(self, widget):
        if self.n_conv_block > 1:
            self.cnn_block_box.remove_row(self.n_conv_block)
            self.n_conv_block -= 1

    def on_callback_selected(self, widget):
        logger.debug("Callback selected: {}".format(widget.get_active_text()))
        if widget.get_active_text().lower() not in self.json_data["callbacks"]:
            self.json_data["callbacks"][widget.get_active_text().lower()] = {}
        if widget.get_active_text().lower() == "modelcheckpoint":
            self.json_data["callbacks"][widget.get_active_text().lower()] = {**{'period':'-1'}, **self.json_data["callbacks"][widget.get_active_text().lower()]}
        elif widget.get_active_text().lower() == "tensorboard":
            self.json_data["callbacks"][widget.get_active_text().lower()] = {**{'training_accuracy':'True', 'training_loss':'True', 'validation_accuracy':'True', 'validation_loss':'True'}, **self.json_data["callbacks"][widget.get_active_text().lower()]}
        elif widget.get_active_text().lower() == "guessingentropy":
            self.json_data["callbacks"][widget.get_active_text().lower()] = {**{'ge_rate':'5', 'r':'0.6', 'n_repeat':'5'}, **self.json_data["callbacks"][widget.get_active_text().lower()]}
        self.update_gridbox_callback(callback_key=widget.get_active_text().lower(), left_gridbox=widget)

    def update_gridbox_callback(self, callback_key, left_gridbox):
        parent_gridbox = left_gridbox.get_parent()
        position = parent_gridbox.query_child(left_gridbox)
        destroy_child = parent_gridbox.get_child_at(position[0]+1, position[1])
        if destroy_child != None:
            parent_gridbox.remove(destroy_child)
        gridbox = gtkGridbox()
        parent_gridbox.attach_next_to(gridbox, left_gridbox, Gtk.PositionType.RIGHT, 1, 1)
        try:
            callback_key = callback_key.lower()
            logger.debug(self.json_data["callbacks"][callback_key])
            for i,p in enumerate(self.json_data["callbacks"][callback_key]):
                gridbox.attach(gtkLabelLeftAlign(label=p), 0, i, 1, 1)
                if self.json_data["callbacks"][callback_key][p] == "True" or self.json_data["callbacks"][callback_key][p] == "False":
                    gridbox.attach(gtkComboText_with_items(["False", "True"], set_active=int(self.json_data["callbacks"][callback_key][p]=="True")), 1, i, 1, 1)
                else:
                    gridbox.attach(gtkResizableEntry(label=self.json_data["callbacks"][callback_key][p]), 1, i, 1, 1)
        except Exception as e:
            logger.debug("Got {}\nError setting callback fields".format(e))

    def on_button_addcallback_clicked(self, widget):
        #insert a new row in the gridbox
        self.callbackbox.insert_row(self.n_callbacks)
        callbackcombo = gtkComboText_with_items(self.callbacks_fileds, set_active=None)
        callbackcombo.connect('changed', self.on_callback_selected)
        self.callbackbox.attach(callbackcombo, 0, self.n_callbacks, 1, 1)
        self.n_callbacks += 1

    def on_button_removecallback_clicked(self, widget):
        if self.n_callbacks > 0:
            self.callbackbox.remove_row(self.n_callbacks-1)
            self.n_callbacks -= 1

    def refresh_listbox_content(self):
        #Remove all Grid elements
        while self.listbox.get_first_child() is not None:
            self.listbox.remove(self.listbox.get_first_child())
        #Add new child boxes
        modelbox = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, spacing=6)
        modelbox.append(gtkLabelLeftAlign(label="Model: "))
        if self.trainbutton.get_active():
            modelcombo = gtkComboText_with_items(self.template_models, set_active=None)
            modelcombo.connect('changed', self.on_model_selected)
            modelbox.append(modelcombo)
            self.listbox.append(modelbox)
            modelcombo.set_active(0)
        elif self.optbutton.get_active():
            modelcombo = gtkComboText_with_items(["MLP", "CNN"], set_active=None)
            modelcombo.connect('changed', self.on_optimize_model_selected)
            modelbox.append(modelcombo)
            self.listbox.append(modelbox)
            modelcombo.set_active(["make_MLP", "make_CNN"].index(self.json_data['hp_search_s']['model_type']) if 'model_type' in self.json_data['hp_search_s'].keys() else 0)
    
    def on_dataset_selected(self, widget):
        logger.debug("Dataset selected: {}".format(widget.get_active_text()))
        self.dataset = widget.get_active_text()
        self.update_gridbox_pyfile(widget.get_parent())

    def on_model_selected(self, widget):
        logger.debug("Model selected: {}".format(widget.get_active_text()))
        
        if "model" not in self.json_data.keys():
            self.json_data['model'] = {}
        if "model_type" in self.json_data['model'].keys():
            try:
                self.json_data['model']["model_type"] = eval(self.json_data['model']["model_type"])
            except Exception as e:
                pass
        if "loss_function" in self.json_data['model'].keys():
            if "crossentropy" in self.json_data['model']["loss_function"].lower():
                self.json_data['model']["loss_function"] = "CrossEntropy"
            elif "mse" in self.json_data['model']["loss_function"].lower():
                self.json_data['model']["loss_function"] = "MSE"
            else:
                self.json_data['model']["loss_function"] = "CrossEntropy"
        if widget.get_active_text() == 'Specific Model':
            self.json_data['model'] = {**{"model_type":"MLP", 
                                        "num_epochs":"100",
                                        "bs":"128",
                                        "loss_function":"CrossEntropy",
                                        "n_layers":1, 
                                        "n_units_l0":"100", 
                                        "activation_function":"ReLu", 
                                        "lr":"0.001", 
                                        "pooling":"Average", 
                                        "optimizer":"Adam", 
                                        "weight_init":"He_uniform"},**self.json_data['model']}
            self.modeltype = self.json_data['model']["model_type"]
        elif widget.get_active_text() == 'MLP':
            self.modeltype = 'MLP'
            self.json_data['model']["model_type"] = "MLP"
            #set default parameters
            self.json_data['model'] = {"model_type":"MLP", 
                                        "num_epochs":"100",
                                        "bs":"128",
                                        "loss_function":"CrossEntropy",
                                        "n_layers":1, 
                                        "n_units_l0":"100", 
                                        "activation_function":"ReLu", 
                                        "lr":"0.001", 
                                        "pooling":"Average", 
                                        "optimizer":"Adam", 
                                        "weight_init":"He_uniform"}
        elif widget.get_active_text() == 'ASCAD_mlp':
            self.modeltype = 'MLP'
            #set default parameters
            self.json_data['model'] = {"model_type":"MLP", 
                                        "num_epochs":"100",
                                        "bs":"128",
                                        "loss_function":"CrossEntropy",
                                        "n_layers":5, 
                                        "n_units_l0":"200",
                                        "n_units_l1":"200",
                                        "n_units_l2":"200",
                                        "n_units_l3":"200", 
                                        "n_units_l4":"200",
                                        "activation_function":"ReLu", 
                                        "lr":"0.00001", 
                                        "pooling":"Average", 
                                        "optimizer":"Adam", 
                                        "weight_init":"He_uniform"}
        elif widget.get_active_text() == 'CNN':
            self.modeltype = 'CNN'
            #set default parameters
            self.json_data['model'] = {"model_type":"CNN", 
                                        "num_epochs":"100",
                                        "bs":"128",
                                        "loss_function":"CrossEntropy",
                                        "n_conv_block":2,
                                        "kernel_size_0":"8",
                                        "c_out_0":"3",
                                        "kernel_size_1":"8",
                                        "c_out_1":"3",
                                        "n_layers":2, 
                                        "n_units_l0":"100",
                                        "n_units_l1":"100",
                                        "activation_function":"ReLu", 
                                        "lr":"0.00001", 
                                        "pooling":"Average", 
                                        "optimizer":"Adam", 
                                        "weight_init":"He_uniform"}
        elif widget.get_active_text() == 'ASCAD_cnn':
            self.modeltype = 'CNN'
            #set default parameters
            self.json_data['model'] = {"model_type":"CNN", 
                                        "num_epochs":"100",
                                        "bs":"128",
                                        "loss_function":"CrossEntropy",
                                        "n_conv_block":5,
                                        "kernel_size_0":"11",
                                        "c_out_0":"64",
                                        "kernel_size_1":"11",
                                        "c_out_1":"128",
                                        "kernel_size_2":"11",
                                        "c_out_2":"256",
                                        "kernel_size_3":"11",
                                        "c_out_3":"512",
                                        "kernel_size_4":"11",
                                        "c_out_4":"512",
                                        "n_layers":2, 
                                        "n_units_l0":"4096",
                                        "n_units_l1":"4096",
                                        "activation_function":"ReLu", 
                                        "lr":"0.001", 
                                        "pooling":"Average", 
                                        "optimizer":"RMSprop", 
                                        "weight_init":"He_uniform"}

        #Remove all child boxes related to model parameters if any
        while len(self.listbox.observe_children()) > 1:
            self.listbox.remove(self.listbox.get_last_child())
        #Add new child boxes related to model parameters
        self.model_parambox = Gtk.Box.new(Gtk.Orientation.VERTICAL, spacing=6)
        self.layerbox = gtkGridbox()
        #cnn parameters
        lcol = 0
        lrow = 0
        self.n_conv_block = int(self.json_data['model']["n_conv_block"]) if "n_conv_block" in self.json_data['model'].keys() else 0
        if self.n_conv_block > 0:
            self.cnn_block_box = gtkGridbox()
            row = 0
            self.cnn_block_box.attach(Gtk.Label(label="kernel_size"), 1, row, 1, 1)
            self.cnn_block_box.attach(Gtk.Label(label="output_size"), 2, row, 1, 1)
            row += 1
            for i in range(self.n_conv_block):
                self.cnn_block_box.attach(Gtk.Label(label="Conv {}".format(row)), 0, row, 1, 1)
                self.cnn_block_box.attach(Gtk.Entry.new_with_buffer(Gtk.EntryBuffer.new(self.json_data['model']["kernel_size_{}".format(i)],-1)), 1, row, 1, 1)
                self.cnn_block_box.attach(Gtk.Entry.new_with_buffer(Gtk.EntryBuffer.new(self.json_data['model']["c_out_{}".format(i)],-1)), 2, row, 1, 1)
                row += 1
            ad_remove_gridbox = gtkAddRemoveGridbox(self.on_button_addconv_clicked, self.on_button_removeconv_clicked)
            self.cnn_block_box.attach(ad_remove_gridbox, 0, row, 1, 1)
            self.layerbox.attach(self.cnn_block_box, lcol, lrow, 1, 1)
            lcol += 1
        #mlp parameters
        self.mlp_layers_box = gtkGridbox()
        row = 0
        self.json_data['model']['n_layers'] = int(self.json_data['model']['n_layers'])
        for i in range(self.json_data['model']['n_layers']):
            self.mlp_layers_box.attach(Gtk.Label(label="Layer {}".format(i+1)), 0, i, 1, 1)
            self.mlp_layers_box.attach(Gtk.Entry.new_with_buffer(Gtk.EntryBuffer.new(self.json_data['model']["n_units_l{}".format(i)],-1)), 1, i, 1, 1)
            row += 1
        ad_remove_gridbox = gtkGridbox()
        ad_remove_gridbox = gtkAddRemoveGridbox(self.on_button_addlayer_clicked, self.on_button_removelayer_clicked)
        self.mlp_layers_box.attach(ad_remove_gridbox, 0, row, 1, 1)
        self.layerbox.attach(self.mlp_layers_box, lcol, lrow, 1, 1)
        lrow += 1
        lcol = 0
        #Other parameters
        self.hparambox = gtkGridbox()
        prow = 0
        self.hparambox.attach(gtkLabelLeftAlign(label="num_epochs"), 0, prow, 1, 1)
        self.hparambox.attach(Gtk.Entry.new_with_buffer(Gtk.EntryBuffer.new(self.json_data['model']['num_epochs'],-1)), 1, prow, 1, 1)
        prow += 1
        self.hparambox.attach(gtkLabelLeftAlign(label="bs"), 0, prow, 1, 1)
        self.hparambox.attach(Gtk.Entry.new_with_buffer(Gtk.EntryBuffer.new(self.json_data['model']['bs'],-1)), 1, prow, 1, 1)
        prow += 1
        self.hparambox.attach(gtkLabelLeftAlign(label="loss_function"), 0, prow, 1, 1)
        combo_button = gtkComboText_with_items(['CrossEntropy', 'MSE'], set_active=[x.lower() for x in ['CrossEntropy', 'MSE']].index(self.json_data['model']['loss_function'].lower()))
        self.hparambox.attach(combo_button, 1, prow, 1, 1)
        prow += 1
        self.hparambox.attach(gtkLabelLeftAlign(label="lr"), 0, prow, 1, 1)
        self.hparambox.attach(Gtk.Entry.new_with_buffer(Gtk.EntryBuffer.new(self.json_data['model']['lr'],-1)), 1, prow, 1, 1)
        prow += 1
        self.hparambox.attach(gtkLabelLeftAlign(label="activation_function"), 0, prow, 1, 1)
        combo_button = gtkComboText_with_items(['ReLu', 'Tanh', 'SeLu'], set_active=[x.lower() for x in ['ReLu', 'Tanh', 'SeLu']].index(self.json_data['model']['activation_function'].strip('"').strip("'").lower()))
        self.hparambox.attach(combo_button, 1, prow, 1, 1)
        prow += 1
        self.hparambox.attach(gtkLabelLeftAlign(label="pooling"), 0, prow, 1, 1)
        combo_button = gtkComboText_with_items(['Average', 'Max'], set_active=[x.lower() for x in ['Average', 'Max']].index(self.json_data['model']['pooling'].strip('"').strip("'").lower()))
        self.hparambox.attach(combo_button, 1, prow, 1, 1)
        prow += 1
        self.hparambox.attach(gtkLabelLeftAlign(label="optimizer"), 0, prow, 1, 1)
        combo_button = gtkComboText_with_items(['Adam', 'SGD', 'RMSprop'], set_active=[x.lower() for x in ['Adam', 'SGD', 'RMSprop']].index(self.json_data['model']['optimizer'].strip('"').strip("'").lower()))
        self.hparambox.attach(combo_button, 1, prow, 1, 1)
        prow += 1
        self.hparambox.attach(gtkLabelLeftAlign(label="weight_init"), 0, prow, 1, 1)
        combo_button = gtkComboText_with_items(['He_uniform', 'Zero'], set_active=[x.lower() for x in ['He_uniform', 'Zero']].index(self.json_data['model']['weight_init'].strip('"').strip("'").lower()))
        self.hparambox.attach(combo_button, 1, prow, 1, 1)
        self.layerbox.attach(self.hparambox, lcol, lrow, 1, 1)
        self.model_parambox.append(self.layerbox)
        self.listbox.append(self.model_parambox)
        logger.debug("json_data['model']: {}".format(self.json_data['model']))

    def on_optimize_model_selected(self, widget):
        logger.debug("Model selected: {}".format(widget.get_active_text()))

        #TODO: if hp_search_s is defined by user in dataset file, load user parameters
        #TODO: remove the unnessary parameters from json_data['hp_search_s'] if user change model type (ex: n_conv_block_s if model_type is MLP), caution, some are added inside Optuna_models.py
        #TODO: add fields for all hparameters 
        if "hp_search_s" not in self.json_data.keys():
            self.json_data['hp_search_s'] = {}
        else:
            for k in self.json_data['hp_search_s']:
                try:
                    self.json_data['hp_search_s'][k] = eval(self.json_data['hp_search_s'][k])
                except Exception as e:
                    pass
        if widget.get_active_text() == 'MLP':
            self.modeltype = 'make_MLP'
            #set default parameters
            self.json_data['hp_search_s'] = {**{"n_layers_s":[1, 5, 1],
                                        "n_neurons_s":[10, 500, 10], 
                                        "activation_function_s":["ReLu", "Tanh", "SeLu"], 
                                        "lr_s":[0.001, 0.1, "log"],
                                        "pooling":"Average", 
                                        "optimizer_s":["Adam"], 
                                        "weight_init":"He_uniform"}, **self.json_data['hp_search_s']}
            self.json_data['hp_search_s']["model_type"] = "make_MLP"
        if widget.get_active_text() == 'CNN':
            self.modeltype = 'make_CNN'
            #set default parameters
            self.json_data['hp_search_s'] = {**{"n_conv_block_s":[1, 5, 1],
                                        "kernel_size_s":[3, 10, 2],
                                        "n_layers_s":[1, 5, 1],
                                        "n_neurons_s":[10, 500, 10], 
                                        "activation_function_s":["ReLu", "Tanh", "SeLu"], 
                                        "lr_s":[0.001, 0.1, "log"],
                                        "pooling":"Average", 
                                        "optimizer_s":["Adam"], 
                                        "weight_init":"He_uniform"}, **self.json_data['hp_search_s']}
            self.json_data['hp_search_s']["model_type"] = "make_CNN"
        #Remove all child boxes related to model parameters if any
        if len(self.listbox.observe_children()) > 1:
            self.listbox.remove(self.listbox.get_last_child())
        
        #Add new child boxes related to model parameters
        self.model_parambox = Gtk.Box.new(Gtk.Orientation.VERTICAL, spacing=6)
        self.layerbox = gtkGridbox()
        row = 0
        self.layerbox.attach(Gtk.Label(label="parameter"), 0, row, 1, 1)
        self.layerbox.attach(Gtk.Label(label="min"), 1, row, 1, 1)
        self.layerbox.attach(Gtk.Label(label="max"), 2, row, 1, 1)
        self.layerbox.attach(Gtk.Label(label="inc"), 3, row, 1, 1)
        row += 1
        if self.json_data['hp_search_s']['model_type'] == 'make_CNN':
            self.layerbox.attach(gtkLabelLeftAlign(label="n_conv_block"), 0, row, 1, 1)
            self.layerbox.attach(Gtk.Entry.new_with_buffer(Gtk.EntryBuffer.new(str(self.json_data['hp_search_s']["n_conv_block_s"][0]),-1)), 1, row, 1, 1)
            self.layerbox.attach(Gtk.Entry.new_with_buffer(Gtk.EntryBuffer.new(str(self.json_data['hp_search_s']["n_conv_block_s"][1]),-1)), 2, row, 1, 1)
            self.layerbox.attach(Gtk.Entry.new_with_buffer(Gtk.EntryBuffer.new(str(self.json_data['hp_search_s']["n_conv_block_s"][2]),-1)), 3, row, 1, 1)
            row += 1
            self.layerbox.attach(gtkLabelLeftAlign(label="kernel_size"), 0, row, 1, 1)
            self.layerbox.attach(Gtk.Entry.new_with_buffer(Gtk.EntryBuffer.new(str(self.json_data['hp_search_s']["kernel_size_s"][0]),-1)), 1, row, 1, 1)
            self.layerbox.attach(Gtk.Entry.new_with_buffer(Gtk.EntryBuffer.new(str(self.json_data['hp_search_s']["kernel_size_s"][1]),-1)), 2, row, 1, 1)
            self.layerbox.attach(Gtk.Entry.new_with_buffer(Gtk.EntryBuffer.new(str(self.json_data['hp_search_s']["kernel_size_s"][2]),-1)), 3, row, 1, 1)
            row += 1
        self.layerbox.attach(gtkLabelLeftAlign(label="n_layers"), 0, row, 1, 1)
        self.layerbox.attach(Gtk.Entry.new_with_buffer(Gtk.EntryBuffer.new(str(self.json_data['hp_search_s']["n_layers_s"][0]),-1)), 1, row, 1, 1)
        self.layerbox.attach(Gtk.Entry.new_with_buffer(Gtk.EntryBuffer.new(str(self.json_data['hp_search_s']["n_layers_s"][1]),-1)), 2, row, 1, 1)
        self.layerbox.attach(Gtk.Entry.new_with_buffer(Gtk.EntryBuffer.new(str(self.json_data['hp_search_s']["n_layers_s"][2]),-1)), 3, row, 1, 1)
        row += 1
        self.layerbox.attach(gtkLabelLeftAlign(label="n_neurons"), 0, row, 1, 1)
        self.layerbox.attach(Gtk.Entry.new_with_buffer(Gtk.EntryBuffer.new(str(self.json_data['hp_search_s']["n_neurons_s"][0]),-1)), 1, row, 1, 1)
        self.layerbox.attach(Gtk.Entry.new_with_buffer(Gtk.EntryBuffer.new(str(self.json_data['hp_search_s']["n_neurons_s"][1]),-1)), 2, row, 1, 1)
        self.layerbox.attach(Gtk.Entry.new_with_buffer(Gtk.EntryBuffer.new(str(self.json_data['hp_search_s']["n_neurons_s"][2]),-1)), 3, row, 1, 1)
        row += 1
        self.layerbox.attach(gtkLabelLeftAlign(label="lr"), 0, row, 1, 1)
        self.layerbox.attach(Gtk.Entry.new_with_buffer(Gtk.EntryBuffer.new(str(self.json_data['hp_search_s']["lr_s"][0]),-1)), 1, row, 1, 1)
        self.layerbox.attach(Gtk.Entry.new_with_buffer(Gtk.EntryBuffer.new(str(self.json_data['hp_search_s']["lr_s"][1]),-1)), 2, row, 1, 1)
        self.layerbox.attach(Gtk.Entry.new_with_buffer(Gtk.EntryBuffer.new(str(self.json_data['hp_search_s']["lr_s"][2]),-1)), 3, row, 1, 1)
        row += 1
        self.layerbox.attach(Gtk.Label(label="values comma separated"), 0, row, 4, 1)
        row += 1
        self.layerbox.attach(gtkLabelLeftAlign(label="activation_functions"), 0, row, 1, 1)
        self.layerbox.attach(Gtk.Entry.new_with_buffer(Gtk.EntryBuffer.new(",".join(self.json_data['hp_search_s']["activation_function_s"]),-1)), 1, row, 3, 1)
        row += 1
        self.layerbox.attach(gtkLabelLeftAlign(label="optimizers"), 0, row, 1, 1)
        self.layerbox.attach(Gtk.Entry.new_with_buffer(Gtk.EntryBuffer.new(",".join(self.json_data['hp_search_s']["optimizer_s"]),-1)), 1, row, 3, 1)
        self.model_parambox.append(self.layerbox)
        self.listbox.append(self.model_parambox)
        logger.debug("json_data['hp_search_s']: {}".format(self.json_data["hp_search_s"]))
            
    def save_model_parameters(self):
        if self.script_action == 'train':
            if self.modeltype == 'MLP':
                for i in range(self.json_data['model']['n_layers']):
                    self.json_data['model']['n_units_l{}'.format(i)] = self.mlp_layers_box.get_child_at(1, i).get_buffer().get_text()
                for param in self.mlp_parameters:
                    child_widget = self.hparambox.get_child_at(1, self.mlp_parameters.index(param))
                    if isinstance(child_widget, Gtk.Entry):
                        self.json_data['model'][param] = self.hparambox.get_child_at(1, self.mlp_parameters.index(param)).get_buffer().get_text()
                    elif isinstance(child_widget, Gtk.ComboBoxText):
                        self.json_data['model'][param] = self.hparambox.get_child_at(1, self.mlp_parameters.index(param)).get_active_text()
            elif self.modeltype == 'CNN':
                self.json_data['model']['n_conv_block'] = self.n_conv_block
                for i in range(self.n_conv_block):
                    self.json_data['model']['kernel_size_{}'.format(i)] = self.cnn_block_box.get_child_at(1, i+1).get_buffer().get_text()
                    self.json_data['model']['c_out_{}'.format(i)] = self.cnn_block_box.get_child_at(2, i+1).get_buffer().get_text()
                for i in range(self.json_data['model']['n_layers']):
                    self.json_data['model']['n_units_l{}'.format(i)] = self.mlp_layers_box.get_child_at(1, i).get_buffer().get_text()
                for param in self.cnn_parameters:
                    child_widget = self.hparambox.get_child_at(1, self.cnn_parameters.index(param))
                    if isinstance(child_widget, Gtk.Entry):
                        self.json_data['model'][param] = self.hparambox.get_child_at(1, self.cnn_parameters.index(param)).get_buffer().get_text()
                    elif isinstance(child_widget, Gtk.ComboBoxText):
                        self.json_data['model'][param] = self.hparambox.get_child_at(1, self.cnn_parameters.index(param)).get_active_text()
            else:
                logger.debug("Unknown model type: {}".format(self.modeltype))
            logger.debug("self.json_data['model']: {}".format(self.json_data['model']))
        elif self.script_action == 'optimize':
            if self.modeltype == 'make_MLP':
                self.json_data['hp_search_s']['n_layers_s'] = [self.layerbox.get_child_at(i, 1).get_buffer().get_text() for i in range(1,4)]
                self.json_data['hp_search_s']['n_neurons_s'] = [self.layerbox.get_child_at(i, 2).get_buffer().get_text() for i in range(1,4)]
                self.json_data['hp_search_s']['lr_s'] = [self.layerbox.get_child_at(i, 3).get_buffer().get_text() for i in range(1,4)]
                self.json_data['hp_search_s']['activation_function_s'] = [x for x in self.layerbox.get_child_at(1, 5).get_buffer().get_text().split(',')]
                self.json_data['hp_search_s']['optimizer_s'] = [x for x in self.layerbox.get_child_at(1, 6).get_buffer().get_text().split(',')]
            elif self.modeltype == 'make_CNN':
                self.json_data['hp_search_s']['n_conv_block_s'] = [self.layerbox.get_child_at(i, 1).get_buffer().get_text() for i in range(1,4)]
                self.json_data['hp_search_s']['kernel_size_s'] = [self.layerbox.get_child_at(i, 2).get_buffer().get_text() for i in range(1,4)]
                self.json_data['hp_search_s']['n_layers_s'] = [self.layerbox.get_child_at(i, 3).get_buffer().get_text() for i in range(1,4)]
                self.json_data['hp_search_s']['n_neurons_s'] = [self.layerbox.get_child_at(i, 4).get_buffer().get_text() for i in range(1,4)]
                self.json_data['hp_search_s']['lr_s'] = [self.layerbox.get_child_at(i, 5).get_buffer().get_text() for i in range(1,4)]
                self.json_data['hp_search_s']['activation_function_s'] = [x for x in self.layerbox.get_child_at(1, 7).get_buffer().get_text().split(',')]
                self.json_data['hp_search_s']['optimizer_s'] = [x for x in self.layerbox.get_child_at(1, 8).get_buffer().get_text().split(',')]
            else:
                logger.debug("Unknown model type: {}".format(self.modeltype))
            logger.debug("self.json_data['hp_search_s']: {}".format(self.json_data['hp_search_s']))
        logger.debug(self.script_action)
        try:
            with open(self.json_filename, 'w') as outfile:
                json.dump(self.json_data, outfile)
        except Exception as e:
            logger.debug("Error writing json file: {}".format(e))

    def update_gridbox_pyfile(self, gridbox):
        #Parse the existing python file and convert to json format that can be reused later
        ds_filename = os.path.join(os.path.dirname(__file__), "datasets/" + str.lower(self.dataset) + ".py")
        self.json_filename = os.path.join(os.getenv('LOCALAPPDATA'),"{}.json".format(str.lower(self.dataset))) if os.name == 'nt' else "/tmp/escapydl/{}.json".format(str.lower(self.dataset))
        os.makedirs(os.path.dirname(self.json_filename), exist_ok=True)
        self.json_data = {}
        try:
            self.json_data = to_json(ds_filename, self.json_filename)
            self.use_validation_dataset = True if "validation_name" in self.json_data['parameters'] else False
            logger.debug(json.dumps(self.json_data, indent=4))
            row = 1
            for p in self.json_data['parameters']:
                if p in self.parameter_fields:
                    if p == "dataset_name":
                        gridbox.get_child_at(2, row).set_file_path(self.json_data['parameters'][p])
                        try:
                            self.json_data['parameters'][p] = eval(self.json_data['parameters'][p])
                        except ValueError:
                            pass
                        if self.use_validation_dataset:
                            gridbox.get_child_at(1, row).set_buffer(Gtk.EntryBuffer(text=self.json_data['parameters'][p]))
                            gridbox.get_child_at(1, row).on_entry_changed(None)#force resize
                            # Check the box will create a new row for a validation dataset field
                            gridbox.get_child_at(3, row).set_active(0)
                            row += 1
                            gridbox.get_child_at(2, row).set_file_path(self.json_data['parameters']["validation_name"])
                            try:
                                self.json_data['parameters']["validation_name"] = eval(self.json_data['parameters']["validation_name"])
                            except ValueError:
                                pass
                        else:
                            gridbox.get_child_at(3, row).set_active(1)
                    gridbox.get_child_at(1, row).set_buffer(Gtk.EntryBuffer(text=self.json_data['parameters'][p]))
                    gridbox.get_child_at(1, row).on_entry_changed(None)#force resize
                    row += 1
        except Exception as e:
            self.json_data = {}
            logger.debug("Got {}\nError reading json file, set all fields to empty".format(e))
            for p in self.parameter_fields:
                gridbox.get_child_at(1, self.parameter_fields.index(p)+1).set_buffer(Gtk.EntryBuffer(text=""))
    
    def update_textview_with_code(self, textview, func_name):
        #Read data from json file
        if not "functions" in self.json_data.keys():
            logger.debug("Error reading json file, set all fields to empty, set default code for known functions")
            self.json_data['functions'] = {}
            self.json_data['functions']["get_attack_parameters"] = "def get_attack_parameters(traceset, subkey, index):\n\treturn {'p':, \n\t\t'k':, \n\t\t'subkey':}"
            self.json_data['functions']["get_known_key"] = "def get_known_key(traceset, subkey, index):\n\treturn "
            self.json_data['functions']["intermediate_value"] = "def intermediate_value(traceset, subkey, index):\n\treturn "
        else:
            logger.debug(self.json_data['functions'])
        for p in self.json_data['functions']:
            if p == func_name:
                textview.buffer.set_text(text=self.json_data['functions'][p])

    def to_first_window(self, widget):
        self.outerbox = gtkVerticalBox()
        hbox = gtkHorizontalBox()
        #Set gridbox for parameters
        self.paramgridbox = gtkGridbox()
        self.paramgridbox.attach(gtkLabelWithTooltip("Dataset", "Select a dataset among the list of available datasets, or select Custom to enter your own dataset"), 0, 0, 1, 1)
        datasets = sorted([f[:-3] for f in os.listdir(os.path.join(os.path.dirname(__file__), "datasets")) if f.endswith(".py") and f!="__init__.py" and f!="to_json.py"])
        combo = gtkComboText_with_items(datasets, set_active=None)
        combo.append_text('Custom')
        self.paramgridbox.attach(combo, 1, 0, 1, 1)
        combo.connect('changed', self.on_dataset_selected)
        for i, f in enumerate(self.parameter_fields):
            self.paramgridbox.attach(gtkLabelLeftAlign(label="{}: ".format(f)), 0, i+1, 1, 1)
            entryval = gtkResizableEntry()
            self.paramgridbox.attach(entryval, 1, i+1, 1, 1)
            if f=="dataset_name":
                self.paramgridbox.attach(gtkButtonFileSelection(entry=entryval), 2, i+1, 1, 1)
                val_dataset_checkbutton = Gtk.CheckButton.new_with_label("use for validation")
                val_dataset_checkbutton.row_index = i+1 # Save the row index to later create or remove validation dataset row below it
                val_dataset_checkbutton.set_active(1) # Set to True by default
                val_dataset_checkbutton.connect("toggled", self.on_button_validationdataset_clicked)
                self.paramgridbox.attach(val_dataset_checkbutton, 3, i+1, 1, 1)
        hbox.append(self.paramgridbox)
        combo.set_active(0)
        self.outerbox.append(hbox)
        
        self.add_navigation_bar(self.to_first_window, self.to_second_window)

        self.window.set_child(self.outerbox)
        self.window.present()

    def to_second_window(self, widget):
        #save previous parameters set in first window to json file
        self.json_data['parameters'] = {}
        parameter_fields = self.parameter_fields_wval if self.use_validation_dataset else self.parameter_fields
        for i, f in enumerate(parameter_fields):
            self.json_data['parameters'][f] = self.paramgridbox.get_child_at(1, i+1).get_buffer().get_text()
        try:
            with open(self.json_filename, 'w') as outfile:
                json.dump(self.json_data, outfile)
        except Exception as e:
            logger.debug("Error writing json file: {}".format(e))
        logger.debug(json.dumps(self.json_data, indent=4))

        
        #set up second window
        self.outerbox = gtkVerticalBox()
        vbox = gtkVerticalBox()

        self.funcgridbox = gtkGridbox()
        #Set a enrtry block to enter code
        self.funcgridbox.attach(gtkLabelLeftAlign(label="get_attack_parameters :"), 0, 0, 1, 1)
        codebox = gtkCodebox()
        self.update_textview_with_code(codebox.textview, "get_attack_parameters")
        self.funcgridbox.attach(codebox, 0, 1, 1, 1)
        #Set a enrtry block to enter code
        self.funcgridbox.attach(gtkLabelLeftAlign(label="get_known_key :"), 0, 2, 1, 1)
        codebox = gtkCodebox()
        self.update_textview_with_code(codebox.textview, "get_known_key")
        self.funcgridbox.attach(codebox, 0, 3, 1, 1)
        #Set a enrtry block to enter code
        self.funcgridbox.attach(gtkLabelLeftAlign(label="intermediate_value :"), 0, 4, 1, 1)
        codebox = gtkCodebox()
        self.update_textview_with_code(codebox.textview, "intermediate_value")
        self.funcgridbox.attach(codebox, 0, 5, 1, 1)
        vbox.append(self.funcgridbox)
        self.outerbox.append(vbox)

        self.add_navigation_bar(self.to_first_window, self.to_third_window, fill_gap=False)

        self.window.set_child(self.outerbox)
        self.window.present()

    def to_third_window(self, widget):
        #save previous window parameters to json file
        for i, f in enumerate(self.functions_fields):
            textbuffer = self.funcgridbox.get_child_at(0, 2*i+1).textview.buffer
            text = textbuffer.get_text(textbuffer.get_start_iter(), textbuffer.get_end_iter(), False)
            self.json_data['functions'][f] = text
        try:
            with open(self.json_filename, 'w') as outfile:
                json.dump(self.json_data, outfile)
        except Exception as e:
            logger.debug("Error writing json file: {}".format(e))
        self.dataset_model = self.json_data['model']

        #set up third window
        self.outerbox = gtkVerticalBox()
        hbox = gtkHorizontalBox()

        self.callbackbox = gtkGridbox()
        self.callbackbox.attach(Gtk.Label(label="Callbacks"), 0, 0, 1, 1)
        row = 1
        #Fill with dataset loaded callbacks parameters
        for f in self.json_data["callbacks"].keys():
            #Set gridbox for parameters
            combo = Gtk.ComboBoxText()
            combo.append_text('ModelCheckpoint')
            combo.append_text('Tensorboard')
            combo.append_text('GuessingEntropy')
            self.callbackbox.attach(combo, 0, row, 1, 1)
            combo.connect('changed', self.on_callback_selected)
            callback_params = Gtk.Grid()
            self.callbackbox.attach(callback_params, 1, row, 1, 1)
            combo.set_active([x.lower() for x in ['ModelCheckpoint', 'Tensorboard', 'GuessingEntropy']].index(f.lower()))
            row += 1
        if len(self.json_data["callbacks"].keys()) == 0:
            #Set gridbox for parameters
            combo = Gtk.ComboBoxText()
            combo.append_text('ModelCheckpoint')
            combo.append_text('Tensorboard')
            combo.append_text('GuessingEntropy')
            self.callbackbox.attach(combo, 0, row, 1, 1)
            combo.connect('changed', self.on_callback_selected)
            callback_params = Gtk.Grid()
            self.callbackbox.attach(callback_params, 1, row, 1, 1)
            row += 1
        #Add +/- buttons
        self.n_callbacks = row
        row += 1
        ad_remove_gridbox = gtkGridbox()
        ad_remove_gridbox = gtkAddRemoveGridbox(self.on_button_addcallback_clicked, self.on_button_removecallback_clicked)
        self.callbackbox.attach(ad_remove_gridbox, 0, row, 1, 1)
        
        
        hbox.append(self.callbackbox)
        self.outerbox.append(hbox)
        
        self.add_navigation_bar(self.to_second_window, self.to_fourth_window)

        self.window.set_child(self.outerbox)
        self.window.present()

    def to_fourth_window(self, widget):
        #save previous window parameters to json file
        self.json_data["callbacks"] = {}
        for i in range(self.n_callbacks-1):
            callback_name = self.callbackbox.get_child_at(0, i+1).get_active_text().lower()
            if callback_name is not None:
                self.json_data["callbacks"][callback_name] = {}
                name_widget, value_widget = 0, 0
                j = 0
                name_widget = self.callbackbox.get_child_at(1, i+1).get_child_at(0, j)
                value_widget = self.callbackbox.get_child_at(1, i+1).get_child_at(1, j)
                while name_widget != None and value_widget != None:
                    param_name = name_widget.get_label()
                    if isinstance(value_widget, Gtk.ComboBoxText):
                        param_value = value_widget.get_active_text()
                    else:
                        param_value = value_widget.get_buffer().get_text()
                    self.json_data["callbacks"][callback_name][param_name] = param_value
                    j += 1
                    name_widget = self.callbackbox.get_child_at(1, i+1).get_child_at(0, j)
                    value_widget = self.callbackbox.get_child_at(1, i+1).get_child_at(1, j)
        try:
            with open(self.json_filename, 'w') as outfile:
                json.dump(self.json_data, outfile)
        except Exception as e:
            logger.debug("Error writing json file: {}".format(e))
        
        #set up fourth window
        self.outerbox = gtkVerticalBox()
        vbox = gtkVerticalBox()
        
        #define to do train or optimize with two check box marks muttually exlusives
        hbox = gtkHorizontalBox()
        self.trainbutton = Gtk.CheckButton.new_with_label("Train")
        self.optbutton = Gtk.CheckButton.new_with_label("Optimize")
        self.trainbutton.connect("toggled", self.on_button_trainopt_clicked)
        self.optbutton.connect("toggled", self.on_button_trainopt_clicked)
        self.trainbutton.set_group(self.optbutton)
        hbox.append(self.trainbutton)
        hbox.append(self.optbutton)
        vbox.append(hbox)
        vbox.append(Gtk.Separator.new(Gtk.Orientation.HORIZONTAL))

        #Contains the model parameters information, refreshed on actions from buttons
        self.listbox = gtkVerticalBox()
        
        #Append the box for model parameters to the main frame
        vbox.append(self.listbox)
        self.outerbox.append(vbox)

        self.add_navigation_bar(self.to_third_window, self.to_fifth_window)

        #Render the window again
        self.window.set_child(self.outerbox)
        self.trainbutton.set_active(self.script_action == 'train')
        self.optbutton.set_active(self.script_action == 'optimize')
        self.window.present()

    def to_fifth_window(self, widget):
        #save model parameters to json file
        self.save_model_parameters()

        #set up fourth window
        self.outerbox = gtkVerticalBox()
        vbox = gtkVerticalBox()

        #Create a ready label
        label = gtkLabelLeftAlign(label="Ready to run")
        label.set_halign(Gtk.Align.START)
        vbox.append(label)

        #Make a run and stop button
        #TODO: when the python script crashes, toggle the run button off
        hbox = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, spacing=6)
        startScriptButton = Gtk.Button.new_with_label("Run")
        startScriptButton.connect("clicked", self.on_button_run_script_clicked)
        hbox.append(startScriptButton)
        stopScriptButton = Gtk.Button.new_with_label("Stop")
        stopScriptButton.connect("clicked", self.on_button_stop_script_clicked)
        hbox.append(stopScriptButton)
        tbButton = Gtk.Button.new_with_label("Tensorboard")
        tbButton.connect("clicked", self.on_button_tensorboarb_clicked)
        hbox.append(tbButton)
        if self.script_action == 'optimize':
            optButton = Gtk.Button.new_with_label("Optuna Dashboard")
            optButton.connect("clicked", self.on_button_optuna_dashboard_clicked)
            hbox.append(optButton)
        vbox.append(hbox)

        #Make a output textview to attach to the script log
        hbox = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, spacing=6)
        hbox.append(gtkLabelLeftAlign(label="Output:"))
        vbox.append(hbox)
        
        #Make a scrolled window to attach to the output textview
        self.output_window = gtkPipeMonitorWindow()

        #Attach the scrolled window to the main frame
        vbox.append(self.output_window)
        self.outerbox.append(vbox)

        self.add_navigation_bar(self.to_fourth_window, self.to_fifth_window, fill_gap=False)

        self.window.set_child(self.outerbox)
        self.window.present()

def main():
    #Parse arguments
    parser = argparse.ArgumentParser(description='eSCAPyDL GUI: Deep Leaning SCA PyTorch framework. Training models for tracesets with SCA metrics.')
    parser.add_argument('-a', '--timestamp', help='Timestamp of the experiment', default=time.strftime("%Y%m%d_%H%M%S",time.localtime()))
    parser.add_argument('-v', '--verbose', help='Verbose level', action='count', default=0)
    parser.add_argument('-l', '--log', help='Activate log to file', default='')
    args = parser.parse_args()
    if args.verbose == 1:
        logger.setLevel(logging.DEBUG)
    app = eSCAPyDLgui(timestamp=args.timestamp, log_filename=args.log, verbose=args.verbose)
    exit_status = app.run()
    sys.exit(exit_status)

if __name__ == "__main__":
    sys.exit(main())


import sys
import json
import re
import logging
logging.basicConfig(format='%(levelname)s:%(message)s')
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# Input file is a python script with a dictionary called parameters and several functions
# The goal is to convert this file into a json file
def to_json(filename, out_filename):
    with open(filename,'r') as f:
        lines = f.readlines()

    #Remove all line comments and block comments
    for i, line in enumerate(lines):
        if line.startswith('#'):
            lines[i] = ""
        elif re.search(r"(.*?)\"\"\"", line) != None:
            lines[i] = ""
            i+=1
            while re.search(r"(.*?)\"\"\"", lines[i]) == None:
                lines[i] = ""
                i+=1
            lines[i] = ""
    #pop empty lines
    for i in range(len(lines)-1,-1,-1):
        if lines[i] == "":
            lines.pop(i)

    #Find extra imports
    extra_imports = []
    for i, line in enumerate(lines):
        if re.search(r"(^import\s.*)", line) != None:
            group_match = re.split(r"(^import\s.*)", line)
            extra_imports.append(group_match[1])
        if re.search(r"(^from\s.*)", line) != None:
            group_match = re.split(r"(^from\s.*)", line)
            extra_imports.append(group_match[1])
    #Find parameters
    parameters = {}
    for i, line in enumerate(lines):
        if re.search(r"(.*?)parameters(.*?)=(.*?)\{", line):
            i+=1
            line = lines[i]
            while not re.search(r"(.*?)\}", line):
                group_match = re.split(r"(\s?)+(\")(.*)(\")(\s+)?(\:)(\s+)?(.*)(\,)",line)
                parameters[group_match[3]] = group_match[8]
                i+=1
                line = lines[i]
    #Find hp_search_s
    hp_search_s = {}
    for i, line in enumerate(lines):
        if re.search(r"(.*?)hp_search_s(.*?)=(.*?)\{", line):
            i+=1
            line = lines[i]
            while not re.search(r"(.*?)\}", line):
                group_match = re.split(r"(\s?)+(\")(.*)(\")(\s+)?(\:)(\s+)?(.*)(\,)",line)
                hp_search_s[group_match[3]] = group_match[8]
                i+=1
                line = lines[i]
    #Find model
    model = {}
    for i, line in enumerate(lines):
        if re.search(r"(.*?)model(.*?)=(.*?)\{", line):
            i+=1
            line = lines[i]
            while not re.search(r"(.*?)\}", line):
                group_match = re.split(r"(\s?)+(\")(.*)(\")(\s+)?(\:)(\s+)?(.*)(\,)",line)
                model[group_match[3]] = group_match[8]
                i+=1
                line = lines[i]
    #Find calkbacks
    callbacks = {}
    for i, line in enumerate(lines):
        if re.search(r"(.*?)callbacks(.*?)=(.*?)\{", line):
            num_open_brackets = line.count('{')
            num_close_brackets = line.count('}')
            while num_open_brackets != num_close_brackets:
                i+=1
                line = lines[i]
                num_open_brackets += line.count('{')
                num_close_brackets += line.count('}')
                while not re.search(r"(.*?)\}", line):
                    if re.search(r"(\s?)+(\")(.*)(\")(\:)(\s?)+(\{)", line): 
                        callback_name = re.split(r"(\s?)+(\")(.*)(\")(\s?)+(\:)(\s?)+(\{)", line)[3]
                        callbacks[callback_name] = {}
                        i+=1
                        line = lines[i]
                        num_open_brackets += line.count('{')
                        num_close_brackets += line.count('}')
                        while not re.search(r"(.*?)\}", line):
                            group_match = re.split(r"(\s?)+(\")(.*)(\")(\s+)?(\:)(\s+)?(.*)(\,)",line)
                            callbacks[callback_name][group_match[3]] = group_match[8]
                            i+=1
                            line = lines[i]
                            num_open_brackets += line.count('{')
                            num_close_brackets += line.count('}')
    #Find functions
    functions = {}
    for i, line in enumerate(lines):
        #If line does not start with a white space or any kind or a newline, save the function name and function description in a dictionary
        if re.search(r"def\s(.*?)\((.*)\)", line) != None:
            group_match = re.split(r"def\s(.*?)\((.*)\)", line)
            function_name = group_match[1]
            functions[function_name] = line
            i+=1
            line = lines[i]
            while not line.startswith('def') and not line.startswith('\n'):
                functions[function_name] += line
                if i == len(lines)-1:
                    break
                i += 1
                line = lines[i]
            functions[function_name] += '\n'
        #If line starts with a variable name
        if re.search(r"(.*)\s?\=\s?(.*)",line) != None:
            group_match = re.split(r"(.*)\s?\=\s?(.*)",line)
            variable_name = group_match[1]
            if not variable_name.startswith('parameters') and not variable_name.startswith('callbacks') and not variable_name.startswith('model') and not variable_name.startswith('hp_search_s'):
                functions[variable_name] = line
                i+=1
                line = lines[i]
                while not line.startswith('def') and not line.startswith('\n'):
                    functions[variable_name] += line
                    if i == len(lines)-1:
                        break
                    i += 1
                    line = lines[i]
                functions[variable_name] += '\n'
    #Save to json
    json_data = {'parameters':parameters,'model':model, 'functions':functions, 'hp_search_s':hp_search_s, 'callbacks':callbacks, 'extra_imports':extra_imports}
    try:
        with open(out_filename,'w') as f:
            json.dump(json_data,f)
        logger.debug('Saved to {}'.format(out_filename))
    except Exception as e:
        logger.debug(e)
    return json_data

if __name__=="__main__":
    filename = sys.argv[1]
    out_filename = sys.argv[2]
    sys.exit(to_json(filename, out_filename))



# eSCAPyDL - a Deep Learning Side-Channel Analysis Python Framework
# Copyright (C) 2023  Weissbart Léo

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os, sys
import time
import numpy as np
import progressbar
import json
import sqlalchemy
import pickle
import ast

import torch
import torch.nn as nn
import torch.nn.functional as F

#---------------------------------------------------#
# Loss functions
class FocalLoss(nn.Module):
    def __init__(
            self,
            weight=None,
            gamma=2.,
            reduction='mean'
    ):
        nn.Module.__init__(self)
        self.weight = weight
        self.gamma = gamma
        self.reduction = reduction

    def forward(self, input_tensor, target_tensor):
        log_prob = F.log_softmax(input_tensor, dim=-1)
        prob = torch.exp(log_prob)
        return F.nll_loss(
                ((1 - prob) ** self.gamma) * log_prob,
                target_tensor,
                weight=self.weight,
                reduction=self.reduction
        )

class RankLoss(nn.Module):
    def __init__(self):
        nn.Module.__init__(self)
    def forward(self, input_tensor, target_tensor):
        log_prob = F.log_softmax(input_tensor, dim=-1)
        prob = torch.exp(log_prob)
        return torch.argwhere(torch.argsort(prob)==target_tensor)

#---------------------------------------------------#
# Custom layers
class LocallyConnected2d(nn.Module):
    """2D Locally connected layer (i.e., convolution without weight sharing)
    Source:https://discuss.pytorch.org/t/locally-connected-layers/26979/2
    """
    def __init__(self, in_channels, out_channels, output_size, kernel_size, stride, bias=False):
        super(LocallyConnected2d, self).__init__()
        output_size = torch.nn.modules.utils._pair(output_size)
        self.weight = nn.Parameter(
            torch.randn(1, out_channels, in_channels, output_size[0], output_size[1], kernel_size**2)
        )
        if bias:
            self.bias = nn.Parameter(
                torch.randn(1, out_channels, output_size[0], output_size[1])
            )
        else:
            self.register_parameter('bias', None)
        self.kernel_size = torch.nn.modules.utils._pair(kernel_size)
        self.stride = torch.nn.modules.utils._pair(stride)
        
    def forward(self, x):
        _, c, h, w = x.size()
        kh, kw = self.kernel_size
        dh, dw = self.stride
        x = x.unfold(2, kh, dh).unfold(3, kw, dw)
        x = x.contiguous().view(*x.size()[:-2], -1)
        # Sum in in_channel and kernel_size dims
        out = (x.unsqueeze(1) * self.weight).sum([2, -1])
        if self.bias is not None:
            out += self.bias
        return out

class LocallyConnected1d(nn.Module):
    """1D Locally connected layer (i.e., convolution without weight sharing)
    Adapted from a 2D implementation
    Source:https://discuss.pytorch.org/t/locally-connected-layers/26979/2
    """
    def __init__(self, in_channels, input_features, out_channels, kernel_size, stride, bias=False):
        super(LocallyConnected1d, self).__init__()
        self.output_size = (input_features - (kernel_size - 1)) // stride if stride==1 else (input_features - (kernel_size - 1) - 1) // stride + 1
        self.weight = nn.Parameter(
            torch.randn(1, out_channels, in_channels, self.output_size, kernel_size)
        )
        if bias:
            self.bias = nn.Parameter(
                torch.randn(1, out_channels, self.output_size)
            )
        else:
            self.register_parameter('bias', None)
        self.kernel_size = kernel_size
        self.stride = stride
        
    def forward(self, x):
        x = x.unfold(2, self.kernel_size, self.stride)
        x = x.contiguous().view(*x.size()[:-1], -1)
        # Sum in in_channel and kernel_size dims
        out = (x.unsqueeze(1) * self.weight).sum([2, -1])
        if self.bias is not None:
            out += self.bias
        return out

#---------------------------------------------------#
# Utilities 
def he_uniform_init_weights(m):
    """Initialize weights of a layer with He uniform distribution

    :param m: layer to initialize
    :type m: torch.nn.Module
    """
    if type(m) == nn.Linear:
        nn.init.kaiming_uniform_(m.weight, mode='fan_in', nonlinearity='relu')#same as glorot_uniform
        if m.bias != None:
            m.bias.data.zero_()

def zero_init_weights(m):
    """Initialize weights of a layer with zeros

    :param m: layer to initialize
    :type m: torch.nn.Module
    """
    if type(m) == nn.Linear:
        m.weight.data.zero_()
        if m.bias != None:
            m.bias.data.zero_()

class NumpyArrayEncoder(json.JSONEncoder):
    """Encode numpy arrays to json

    :param json: json encoder
    :type json: json.JSONEncoder
    """
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)
    
def save_history(history,filename):
    """Save training history to json file

    :param history: training history of the model
    :type history: dict
    :param filename: name of the file to save
    :type filename: str
    """
    with open(filename.replace('.json','').rstrip('/')+'.json','w') as output:
        json.dump(history,output,cls=NumpyArrayEncoder)
    return

def load_history(filename):
    """Load training history from json file

    :param filename: name of the file to load
    :type filename: str
    """
    with open(filename,'r') as f:
        ret = json.load(f)
    return ret

def save_model_weights(model,filename):
    """Save the weights of a model to a file

    :param model: model to save
    :type model: torch.nn.Module
    :param filename: name of the file to save
    :type filename: str
    """
    torch.save(model.state_dict(), filename)
    return

def load_model_weights(model,filename):
    """Load the weights of a model from a file to model

    :param model: model to load
    :type model: torch.nn.Module
    :param filename: name of the file to load
    :type filename: str
    """
    return model.load_state_dict(torch.load(filename))

def save_model_torchscript(model,filename):
    """Save the torchscript of a model to a file

    :param model: model to save
    :type model: torch.nn.Module
    :param filename: name of the file to save
    :type filename: str
    """
    return torch.jit.save(torch.jit.script(model),filename)

def load_model_torchscript(filename):
    """Load the torchscript of a model from a file to model
    Call model.eval() on returned model if you want to do inference on the model

    :param filename: name of the file to load
    :type filename: str
    """
    return torch.jit.load(filename)

def shuffle_numpy(x):
    """Shuffle a numpy array

    :param x: array to shuffle
    :type x: numpy.ndarray
    """
    p = np.random.permutation(x[0].shape[0])
    for i in range(len(x)):
        x[i] = x[i][p]

def shuffle_dict(x,shuffled_keys):
    """Shuffle a dictionary of arrays

    :param x: dictionary of arrays to shuffle
    :type x: dict
    :param shuffled_keys: list of keys to shuffle
    :type shuffled_keys: list
    """
    p = np.random.permutation(x[shuffled_keys[0]].shape[0])
    for i in shuffled_keys:
        x[i] = x[i][p]

def shuffle_torch(x):
    """Shuffle a torch tensor

    :param x: tensor to shuffle
    :type x: torch.Tensor
    """
    p = torch.randperm(x[0].shape[0])
    for i in range(len(x)):
        x[i] = x[i][p]
    return x

def save_db(db_filepath, dataset_name, dataset_parameters, model, history):
    """Save the dataset parameters, model, and training history to a sqlite database
    Uses sqlalchemy to sqlite database

    Structure of the database:\n
    - alembic_version\n
    - datset_parameters (id, dataset_name, value)\n
    - model (id, model_name, model_hyperparameters, model_class, fit_function, pickled_model_weights)\n
    - training_outputs (id, training_history)

    :param db_filepath: path to the database file
    :type db_filepath: str
    :param dataset_parameters: Dictionary of dataset parameters
    :type dataset_parameters: dict
    :param model: model to save
    :type model: torch.nn.Module
    :param history: training history of the model
    :type history: dict
    """
    engine = sqlalchemy.create_engine('sqlite:///'+db_filepath, echo=True)
    con = engine.connect()
    con.execute(sqlalchemy.text("CREATE TABLE alembic_version (version_num VARCHAR(32) NOT NULL, CONSTRAINT alembic_version_pkc PRIMARY KEY (version_num))"))
    con.execute(sqlalchemy.text("INSERT INTO alembic_version VALUES(:version_num)"), [{"version_num":sqlalchemy.__version__}])
    con.commit()

    intermediate_value_code = ""
    get_attack_parameters_code = ""
    get_known_key_code = ""
    with open(os.path.join(os.path.join(os.path.dirname(__file__), "datasets"), dataset_name+".py"), "r") as f:
        callbacks_file = ast.parse(f.read())
        for p in callbacks_file.body:
            if isinstance(p, ast.FunctionDef):
                if p.name == "intermediate_value":
                    intermediate_value_code = ast.unparse(p)
                if p.name == "get_attack_parameters":
                    get_attack_parameters_code = ast.unparse(p)
                if p.name == "get_known_key":
                    get_known_key_code = ast.unparse(p)
    con.execute(sqlalchemy.text("CREATE TABLE IF NOT EXISTS datset_parameters (id PRIMARY KEY, dataset_name TEXT, dataset_parameters JSON, intermediate_value_code TEXT, get_attack_parameters_code TEXT, get_known_key_code TEXT)"))
    con.execute(
        sqlalchemy.text("INSERT INTO datset_parameters VALUES(1, :dataset_name, :dataset_parameters, :intermediate_value_code, :get_attack_parameters_code, :get_known_key_code)"),
        [{"dataset_name":dataset_name,
          "dataset_parameters":json.dumps(dataset_parameters),
          "intermediate_value_code":intermediate_value_code,
          "get_attack_parameters_code":get_attack_parameters_code,
          "get_known_key_code":get_known_key_code,}],
    )

    model.Hparams['loss_function'] = str(model.Hparams['loss_function'])
    model_code = ""
    fit_function_code = ""
    with open(os.path.join(os.path.dirname(__file__), "Models.py"), "r") as f:
        models_file = ast.parse(f.read())
        for p in models_file.body:
            if isinstance(p, ast.ClassDef):
                if p.name == model.__class__.__name__:
                    model_code = ast.unparse(p)
            if isinstance(p, ast.FunctionDef):
                if p.name == model.fit.__name__:
                    fit_function_code = ast.unparse(p)
    con.execute(sqlalchemy.text("CREATE TABLE IF NOT EXISTS model (id PRIMARY KEY, model_name TEXT, model_hyperparameters TEXT, model_class TEXT, fit_function TEXT, pickled_model_weights BLOB)"))
    con.execute(
        sqlalchemy.text("INSERT INTO model VALUES(1, :model_name, :model_hyperparameters, :model_class, :fit_function, :pickled_model_weights)"),
        [{  "model_name":model.name,
            "model_hyperparameters":json.dumps(model.Hparams),
            "model_class":model_code,
            "fit_function":fit_function_code,
            "pickled_model_weights":pickle.dumps(model.state_dict()),
        }]
    )

    con.execute(sqlalchemy.text("CREATE TABLE IF NOT EXISTS training_outputs (id PRIMARY KEY, history JSON)"))
    con.execute(
        sqlalchemy.text("INSERT INTO training_outputs VALUES(1, :history)"),
        [{"history":json.dumps(history,cls=NumpyArrayEncoder)}]
    )
    con.commit()

def load_db(db_filepath):
    """Load the dataset parameters, model, and training history from a sqlite database

    :param db_filepath: path to the database file
    :type db_filepath: str
    """
    engine = sqlalchemy.create_engine('sqlite:///'+db_filepath, echo=True)
    con = engine.connect()
    
    result = con.execute(sqlalchemy.text("SELECT * FROM datset_parameters")).first()
    dataset_name = result[1]
    dataset_parameters = result[2:]
    print("dataset_name = ",dataset_name)
    print("dataset_parameters = ", dataset_parameters)

    result = con.execute(sqlalchemy.text("SELECT model_hyperparameters, model_class, fit_function, pickled_model_weights FROM model")).first()
    model_hyperparameters = json.loads(result[0])
    model_class = result[1]
    fit_function = result[2]
    pickled_model_weights = pickle.loads(result[3])
    print(model_hyperparameters, model_class, fit_function, pickled_model_weights)

    result = con.execute(sqlalchemy.text("SELECT history FROM training_outputs")).first()
    training_history = json.loads(result[0])
    print(training_history)
    return dataset_name


def pbar(iter,desc=''):
    widgets = [
        progressbar.Bar(marker='█'),
        progressbar.Counter(format='%(value)d/%(max_value)d'),
        ' [', progressbar.Timer(format='%(elapsed)s'), '<', progressbar.ETA(format='%(eta)8s', format_not_started='--:--:--', format_finished='%(elapsed)8s', format_zero='00:00:00'), '] ',
    ]
    bar = progressbar.ProgressBar(redirect_stdout=True, widgets=widgets, suffix='\\r', prefix=desc, max_value=len(iter)).start()
    for i,it in enumerate(iter):
        yield it
        bar.update(i+1)
    bar.finish(end='\n')




import torch
import sys,os
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../escapydl/')))

traceset = {
    "plaintext" : torch.randint(0, 256, (2,16), dtype=torch.uint8),
    "key"       : torch.randint(0, 256, (2,16), dtype=torch.uint8),
    "offset"    : torch.randint(0, 256, (2,16), dtype=torch.uint8),
}
subkey = 0
index = 0

## Test all dataset functions

## AESRD
def test_aesrd_get_attack_parameters():
    from escapydl.datasets.aesrd import get_attack_parameters
    output = get_attack_parameters(traceset, subkey, index)
    assert output == {'p': traceset["plaintext"][index,subkey], 'k': traceset["key"][index,subkey], 'subkey':subkey}, "expected output: {}, actual output: {}".format({'p': traceset["plaintext"][index,subkey], 'k': traceset["key"][index,subkey], 'subkey':subkey}, output)

def test_aesrd_get_known_key():
    from escapydl.datasets.aesrd import get_known_key
    output = torch.Tensor(get_known_key(traceset, subkey, index))
    assert output == torch.Tensor(traceset["key"][index,subkey]), "expected output: {}, actual output: {}".format(torch.Tensor(traceset["key"][index,subkey]), output)

def test_aesrd_intermediate_value():
    from escapydl.datasets.aesrd import intermediate_value, get_attack_parameters, get_known_key
    output = intermediate_value(get_known_key(traceset, subkey, index), get_attack_parameters(traceset, subkey, index))
    assert output.shape == (1,), "expected output: {}, actual output: {}".format((1,), output.shape)

## ASCAD_fixed
def test_ascad_fixed_get_attack_parameters():
    from escapydl.datasets.ascad_fixed import get_attack_parameters
    output = get_attack_parameters(traceset, subkey, index)
    assert output == {'p': traceset["plaintext"][index,subkey], 'k': traceset["key"][index,subkey], 'subkey':subkey}, "expected output: {}, actual output: {}".format({'p': traceset["plaintext"][index,subkey], 'k': traceset["key"][index,subkey], 'subkey':subkey}, output)

def test_ascad_fixed_get_known_key():
    from escapydl.datasets.ascad_fixed import get_known_key
    output = torch.Tensor(get_known_key(traceset, subkey, index))
    assert output == torch.Tensor(traceset["key"][index,subkey]), "expected output: {}, actual output: {}".format(torch.Tensor(traceset["key"][index,subkey]), output)

def test_ascad_fixed_intermediate_value():
    from escapydl.datasets.ascad_fixed import intermediate_value, get_attack_parameters, get_known_key
    output = intermediate_value(get_known_key(traceset, subkey, index), get_attack_parameters(traceset, subkey, index))
    assert output.shape == (1,), "expected output: {}, actual output: {}".format((1,), output.shape)

## ASCAD_random
def test_ascad_random_get_attack_parameters():
    from escapydl.datasets.ascad_random import get_attack_parameters
    output = get_attack_parameters(traceset, subkey, index)
    assert output == {'p': traceset["plaintext"][index,subkey], 'k': traceset["key"][index,subkey], 'subkey':subkey}, "expected output: {}, actual output: {}".format({'p': traceset["plaintext"][index,subkey], 'k': traceset["key"][index,subkey], 'subkey':subkey}, output)

def test_ascad_random_get_known_key():
    from escapydl.datasets.ascad_random import get_known_key
    output = torch.Tensor(get_known_key(traceset, subkey, index))
    assert output == torch.Tensor(traceset["key"][index,subkey]), "expected output: {}, actual output: {}".format(torch.Tensor(traceset["key"][index,subkey]), output)

def test_ascad_random_intermediate_value():
    from escapydl.datasets.ascad_random import intermediate_value, get_attack_parameters, get_known_key
    output = intermediate_value(get_known_key(traceset, subkey, index), get_attack_parameters(traceset, subkey, index))
    assert output.shape == (1,), "expected output: {}, actual output: {}".format((1,), output.shape)

## DPAv4
def test_dpav4_get_attack_parameters():
    from escapydl.datasets.dpav4 import get_attack_parameters
    output = get_attack_parameters(traceset, subkey, index)
    assert output == {'p': traceset["plaintext"][index,subkey], 'k': traceset["key"][index,subkey], 'o':traceset["offset"][index,subkey], 'subkey':subkey}, "expected output: {}, actual output: {}".format({'p': traceset["plaintext"][index,subkey], 'k': traceset["key"][index,subkey], 'o':traceset["offset"][index,subkey], 'subkey':subkey}, output)

def test_dpav4_get_known_key():
    from escapydl.datasets.dpav4 import get_known_key
    output = torch.Tensor(get_known_key(traceset, subkey, index))
    assert output == torch.Tensor(traceset["key"][index,subkey]), "expected output: {}, actual output: {}".format(torch.Tensor(traceset["key"][index,subkey]), output)

def test_dpav4_intermediate_value():
    from escapydl.datasets.dpav4 import intermediate_value, get_attack_parameters, get_known_key
    output = intermediate_value(get_known_key(traceset, subkey, index), get_attack_parameters(traceset, subkey, index))
    assert output.shape == (1,), "expected output: {}, actual output: {}".format((1,), output.shape)

if __name__ == "__main__":
    test_aesrd_get_attack_parameters()
    print("test_aesrd_get_attack_parameters passed")

    test_aesrd_get_known_key()
    print("test_aesrd_get_known_key passed")

    test_aesrd_intermediate_value()
    print("test_aesrd_intermediate_value passed")

    test_ascad_fixed_get_attack_parameters()
    print("test_ascad_fixed_get_attack_parameters passed")

    test_ascad_fixed_get_known_key()
    print("test_ascad_fixed_get_known_key passed")

    test_ascad_fixed_intermediate_value()
    print("test_ascad_fixed_intermediate_value passed")

    test_ascad_random_get_attack_parameters()
    print("test_ascad_random_get_attack_parameters passed")

    test_ascad_random_get_known_key()
    print("test_ascad_random_get_known_key passed")

    test_ascad_random_intermediate_value()
    print("test_ascad_random_intermediate_value passed")

    test_dpav4_get_attack_parameters()
    print("test_dpav4_get_attack_parameters passed")

    test_dpav4_get_known_key()
    print("test_dpav4_get_known_key passed")

    test_dpav4_intermediate_value()
    print("test_dpav4_intermediate_value passed")


# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'escapydl'
copyright = '2023, Léo Weissbart'
author = 'Léo Weissbart'
release = '0.0.1'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration
import sys, os
sys.path.insert(0, os.path.abspath('..'))
sys.path.insert(0, os.path.abspath('../..'))
sys.path.insert(0, os.path.abspath('../../escapydl'))

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.viewcode',
    'myst_parser',
]

templates_path = ['_templates']
exclude_patterns = ['*.svg','*.png']



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']



# eSCAPyDL - a Deep Learning Side-Channel Analysis Python Framework
# Copyright (C) 2023  Weissbart Léo

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os, sys
import inspect
import shutil
import json
import logging 
logging.basicConfig(format='%(message)s')
logger = logging.getLogger()
logger.setLevel(logging.INFO)
import time
import argparse

import torch
from torch.utils import tensorboard
import Dataset
import Callbacks
import Optuna_models

import optuna
from optuna.trial import TrialState

def main():
    #Parse arguments
    #TODO: remove default dataset and model and default to help
    parser = argparse.ArgumentParser(description='eSCAPyDL: Deep Leaning SCA PyTorch framework. Training models for tracesets with SCA metrics.')
    parser.add_argument('-d', '--dataset', help='The dataset to use for training. (All datasets have different functions for intermediate value and power model).', default="dpav4")
    parser.add_argument('-m', '--model', help='Type of NN to use (choices: MLP, CNN) CNN is the onetrace implementation.', default="make_MLP")
    parser.add_argument('-s', '--study', help='Optuna study name, for parallelization of study')
    parser.add_argument('-u', '--database', help='Optuna database name', default="sqlite:///cnn_sca.db")
    parser.add_argument('-j', '--json', help='Input data from json file')
    parser.add_argument('-a', '--timestamp', help='Timestamp of the experiment', default=time.strftime("%Y%m%d_%H%M%S",time.localtime()))
    parser.add_argument('-v', '--verbose', help='Verbose level', action='count', default=0)
    parser.add_argument('-l', '--log', help='Log into a file at path', default='')
    args = parser.parse_args()
    __DL_type__ = args.model
    available_models = [x[0] for x in inspect.getmembers(Optuna_models, inspect.isfunction) if x[0].startswith("make_")]
    assert __DL_type__ in available_models, "Model {} not found, available models are: {}".format(__DL_type__, available_models)
    if args.verbose == 1:
        logger.setLevel(logging.DEBUG)
        logging.getLogger("Dataset").setLevel(logging.DEBUG)
        logging.getLogger("Callbacks").setLevel(logging.DEBUG)
        logging.getLogger("optuna_models").setLevel(logging.DEBUG)
    if args.log != '':
        file_handler = logging.FileHandler(args.log, mode='w', encoding='utf-8')
        file_handler.setFormatter(logging.Formatter('%(asctime)s %(levelname)s %(message)s', datefmt='%d/%m/%Y %H:%M:%S'))
        logger.addHandler(file_handler)
    __trs_dataset__ = args.dataset
    db_name = args.database
    study_name = args.study
    if args.json == "None":
        args.json = None
    if args.json != None:
        try:
            with open(args.json, 'r') as f:
                json_data = json.load(f)
        except FileNotFoundError:
            logging.error("Json File '{}' not found".format(args.json))
            exit(1)
    else:
        json_data = {}

    #Main program
    try:
        # Create folders
        timestamp = args.timestamp
        filebase = '{}/{}/{}'.format(__trs_dataset__, __DL_type__, timestamp)
        root                = os.getcwd() + '/experiments/' + filebase + '/'
        os.makedirs(root)
        model_folder        = root + 'models/'
        last_model_folder   = root + 'last_models/'
        history_folder      = root + 'history/'
        result_folder       = root + 'results/'
        for folder in [model_folder,model_folder,last_model_folder,history_folder,result_folder]:
            try:
                os.mkdir(folder)
            except:
                pass

        #load traceset
        dataset = Dataset.Dataset(__trs_dataset__,__DL_type__,json_data=json_data)
        parameters = dataset.get_parameters()
        trace_data = dataset.traceset()
        callbacks_params = dataset.get_callbacks()
        hp_search_s = dataset.get_hp_search_s()

        #Create callbacks
        callbacks = Callbacks.init_callbacks(callbacks_params, 
                                   traceset=trace_data,
                                   parameters=parameters, 
                                   exp_path=result_folder, 
                                   writer_path='runs/'+filebase)

        #Train model
        metric, direction = "gge", "minimize"
        objective = Optuna_models.objective_hp(parameters=parameters,
                            hp_search_s=hp_search_s,
                            trace_data=trace_data,
                            callbacks=callbacks,
                            verbose=1,
                            model_name=__DL_type__,
                            metric=metric
                        )
        pruner = optuna.pruners.PatientPruner(optuna.pruners.MedianPruner(n_startup_trials=5, n_warmup_steps=5, interval_steps=1), patience=30)
        
        if study_name!=None:
            logging.debug("Loading study from name: {}".format(study_name))
            study = optuna.load_study(study_name=study_name, storage=db_name)
        else:
            study_name = "{}_{}_{}".format(__trs_dataset__, __DL_type__, timestamp)
            study = optuna.create_study(sampler=optuna.samplers.TPESampler(), 
                                        direction=direction, 
                                        pruner=pruner, 
                                        study_name=study_name, 
                                        storage=db_name)
        study.optimize(objective, n_trials=100, timeout=None)

        pruned_trials = study.get_trials(deepcopy=False, states=[TrialState.PRUNED])
        complete_trials = study.get_trials(deepcopy=False, states=[TrialState.COMPLETE])

        logging.debug("Study statistics: ")
        logging.debug("  Number of finished trials: ", len(study.trials))
        logging.debug("  Number of pruned trials: ", len(pruned_trials))
        logging.debug("  Number of complete trials: ", len(complete_trials))

        logging.debug("Best trial:")
        trial = study.best_trial

        logging.debug("  Value: ", trial.value)

        logging.debug("  Params: ")
        for key, value in trial.params.items():
            logging.debug("    {}: {}".format(key, value))
    except:
        # Create handler to remove uncessful experiment folder tree if program crashed
        logger.warning("------------------\nException caught. Cleaning experiment folder..\n------------------")
        if os.path.isdir(root):
            shutil.rmtree(root)
        if os.path.isdir('runs/'+filebase):
            shutil.rmtree('runs/'+filebase)
        # optuna.delete_study(study_name=study_name, storage=f"sqlite:///cnn_sca.db")
        logging.exception('')
    finally:
        logging.error("------------------\nEnd of main reached!")
    logging.error("------------------")

if __name__ == "__main__":
    sys.exit(main())


import os
import torch
from helpers import aes_helper as aes

#Dataset parameters
parameters = {
"dataset_name"        : os.getcwd() + '/data/ascad_random.trs',
"dataset_size"        : 300000,
"training_size"       : 200000,
"val_size"            : 50000,
"subkey"              : 2,
"num_key_hypotheses"  : 256,
"n_samples"           : 1400,
"r"                   : 0.6,
"n_repeat"            : 5,
"ge_rate"             : 5,
"shuffle"             : True,
}

model = {
    "num_epochs"          : 100,
    "loss_function"       : 'crossentropy',
    "bs"                  : 128,
    "lr"                  : 1e-3,
}

callbacks = {
    "modelcheckpoint":{
        "period":-1,
    },
    "guessingentropy":{
        "ge_rate":5,
        "n_repeat":5,
        "r":0.6,
    },
    "tensorboard":{
        "training_accuracy":True,
        "validation_accuracy":True,
        "training_loss":True,
        "validation_loss":True,
    },
}

def get_attack_parameters(traceset, subkey, index):
    """Returns a dictionary of parameters to be used to search for attack .
    To be used before calling intermediate_value().

    :param traceset: Traceset to use
    :type traceset: dict
    :param subkey: Subkey to attack
    :type subkey: int
    :param index: Index of the traceset to use
    :type index: int or slice
    :return: Dictionary of parameters
    :rtype: dict
    """
    return {'p':traceset["plaintext"][index,subkey], 
            'k':traceset["key"][index,subkey],
            'subkey':subkey}

def get_known_key(traceset, subkey, index):
    return traceset['key'][index][subkey]

def intermediate_value(keyguess, trace_parameters):
    """Return the intermediate value of a key in the SBox . 
    First xor corresponds to the plaintext blinding, then we do addRoundKey with key guess and apply rsm fo subBytes

    :param trace_parameters: Dictionary containing the trace parameters for intermediate value computation
    :type trace_parameters: dict
    :return: Intermediate value
    :rtype: Any
    """
    if isinstance(trace_parameters['p'],torch.Tensor):
        return torch.index_select(aes.sbox, 0, (trace_parameters['p'] ^ keyguess).int())
    else:
        return aes.sbox.numpy()[trace_parameters['p'] ^ keyguess]


import torch
import numpy as np
import sys,os
import shutil
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

## Test functions in Callbakcs.py
def test_get_attack_parameters():
    from escapydl.Callbacks import get_attack_parameters
    assert get_attack_parameters(0,0,0) == None

def test_get_known_key():
    from escapydl.Callbacks import get_known_key
    assert get_known_key(0,0,0) == None

def test_intermediate_value():
    from escapydl.Callbacks import intermediate_value
    assert intermediate_value(0,0) == None

def test_callback():
    from escapydl.Callbacks import Callback
    callback = Callback()
    assert callback.on_train_begin() == None
    assert callback.on_train_end() == None
    assert callback.on_epoch_begin() == None
    assert callback.on_epoch_end() == None
    assert callback.on_batch_begin() == None
    assert callback.on_batch_end() == None
    assert callback.on_loss_begin() == None
    assert callback.on_loss_end() == None
    assert callback.on_step_begin() == None
    assert callback.on_step_end() == None
    assert callback.reset() == None

def test_ModelCheckpoint():
    from escapydl.Callbacks import ModelCheckpoint
    path = "model_checkpoint_unit_test"
    period = 1
    callback = ModelCheckpoint(path, period=period)
    assert os.path.exists(path) == True, "expected output: {}, actual output: {}".format(True, os.path.exists(path))
    callback.reset()
    model = torch.nn.Linear(10, 10)
    callback.on_epoch_end(model,epoch=1)
    assert os.path.exists(os.path.join(path, 'model_{:03d}epochs.pt'.format(1))) == True, "expected output: {}, actual output: {}".format(True, os.path.exists(os.path.join(path, 'model_{:03d}epochs.pt'.format(0))))
    os.remove(os.path.join(path, 'model_{:03d}epochs.pt'.format(1)))
    callback.on_train_end(model)
    assert os.path.exists(os.path.join(path, 'model.pt')) == True, "expected output: {}, actual output: {}".format(True, os.path.exists(os.path.join(path, 'model.pt')))
    os.remove(os.path.join(path, 'model.pt'))
    os.rmdir(path)

def test_TensorboardCallback():
    from escapydl.Callbacks import TensorboardCallback
    callback = TensorboardCallback()
    callback.reset()
    callback.on_epoch_end()
    #There is nothing to test here, since the callback is empty, but it should pass without errors
    path = "tensorboard_callback_unit_test"
    callback = TensorboardCallback(path=path,callbacks_param={"training_loss":"True", "validation_loss":"True", "training_accuracy":"True", "validation_accuracy":"True"})
    assert os.path.exists(path) == True, "expected output: {}, actual output: {}".format(True, os.path.exists(path))
    callback.reset()
    callback.on_epoch_end(epoch=0, history={"loss":[1], "accuracy":[1], "val_loss":[1], "val_accuracy":[1]})
    assert True
    shutil.rmtree(path)

def test_GuessingEntropyCallback():
    from escapydl.Callbacks import GuessingEntropyCallback
    from escapydl import Callbacks
    Callbacks.get_known_key = lambda traceset, subkey, index: traceset["key"][index]
    Callbacks.intermediate_value = lambda key, attack_parameters: key
    traceset = {
        "x_val" : torch.randn(10, 10),
        "m_val" : {
            "key": torch.ByteTensor([0,0,0,0,0,0,0,0,0,0]),
            "label": torch.randint(0, 2, (10, 1))
        }
    }
    exp_path = "guessing_entropy_callback_unit_test"
    writer_path = "guessing_entropy_callback_unit_test_writter"
    callback = GuessingEntropyCallback(traceset=traceset, 
                                       subkey=[0],
                                       number_of_key_hypothesis=2,
                                       epochs=3,
                                       rate = 1,
                                       n_repeat=1,
                                       r = 0.6,
                                       Hparams={"num_classes":2}, 
                                       exp_path=exp_path, 
                                       writer_path=writer_path)
    assert os.path.exists(exp_path) == True, "expected output: {}, actual output: {}".format(True, os.path.exists(exp_path))
    assert os.path.exists(writer_path) == True, "expected output: {}, actual output: {}".format(True, os.path.exists(writer_path))
    callback.reset()
    assert callback.on_train_begin() == 0
    model = torch.nn.Linear(10, 2)
    callback.on_epoch_end(model, epoch=0)
    callback.on_train_end()
    assert os.path.exists(os.path.join(exp_path, 'GE.npy')) == True, "expected output: {}, actual output: {}".format(True, os.path.exists(os.path.join(exp_path, 'GE.npy')))
    shutil.rmtree(exp_path)
    shutil.rmtree(writer_path)
    fit_value = callback.get_fit_value()
    assert fit_value == 0, "expected output: {}, actual output: {}".format(0, fit_value)
    ge = callback.get_guessing_entropy()
    assert ge.shape == (1,3,4), "expected output: {}, actual output: {}".format((1,3,4), ge.shape)

def test_init_callbacks():
    from escapydl.Callbacks import init_callbacks
    from escapydl import Callbacks
    Callbacks.get_known_key = lambda traceset, subkey, index: traceset["key"][index]
    Callbacks.intermediate_value = lambda key, attack_parameters: key
    traceset = {
        "x_val" : torch.randn(10, 10),
        "m_val" : {
            "key": torch.ByteTensor([0,0,0,0,0,0,0,0,0,0]),
            "label": torch.randint(0, 2, (10, 1))
        }
    }
    Hparams = {"num_classes":2,
               "num_epochs": 3,}
    parameters = {
        "n_samples": 10,
        "num_classes": 2,
        "training_size": 10,
        "val_size": 2,
        "subkey": [0,1],
        "num_key_hypotheses": 2,
    }
    exp_path = "init_callbacks_unit_test"
    writer_path = "init_callbacks_unit_test_writter"
    callbacks_params = {
        "modelcheckpoint": {
            "path": exp_path,
            "period": 1
        },
        "tensorboard": {
            "path": writer_path,
            "callbacks_param": {
                "training_loss": "True", 
                "validation_loss": "True", 
                "training_accuracy": "True", 
                "validation_accuracy": "True"
            }
        },
        "guessingentropy": {
            "ge_rate": 1,
            "n_repeat": 1,
            "r": 0.6,
        }
    }
    callbacks = init_callbacks(callbacks_params, 
                               traceset=traceset, 
                               Hparams=Hparams, 
                               parameters=parameters, 
                               exp_path=exp_path, 
                               writer_path=writer_path)
    assert len(callbacks) == 3, "expected output: {}, actual output: {}".format(3, len(callbacks))
    shutil.rmtree(exp_path)
    shutil.rmtree(writer_path)

if __name__ == "__main__":
    test_get_attack_parameters()
    print("test_get_attack_parameters passed")

    test_get_known_key()
    print("test_get_known_key passed")

    test_intermediate_value()
    print("test_intermediate_value passed")

    test_callback()
    print("test_callback passed")

    test_ModelCheckpoint()
    print("test_ModelCheckpoint passed")

    test_TensorboardCallback()
    print("test_TensorboardCallback passed")

    test_GuessingEntropyCallback()
    print("test_GuessingEntropyCallback passed")

    test_init_callbacks()
    print("test_init_callbacks passed")


# eSCAPyDL - a Deep Learning Side-Channel Analysis Python Framework
# Copyright (C) 2023  Weissbart Léo

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import inspect, types
import sys
import time
import logging 
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
logger.debug(device)

import Utils
import Callbacks

from torch.utils import tensorboard
import optuna
#---------------------------------------------------#
# Hyperparameter optimization

def is_function_local(object):
    return isinstance(object, types.FunctionType) and object.__module__ == __name__

hp_search_s_default = {"activation_function_s": ['relu', 'tanh', 'selu'],
                            "epochs": [100,100,1], 
                            "bs": [32,512,32],
                            "loss_function_s": ["crossentropy"],
                            "n_neurons_s": [10,500,10], 
                            "n_layers_s": [1,5,1], 
                            "kernel_size_s": [3,8,1],
                            "n_conv_block_s": [0,2,1],
                            "optimizer_s": ["Adam"], 
                            "lr_s": [1e-3,1e-3,'log']}

def make_MLP(trial, hp_search_s, trace_length, num_classes, **kwargs):
    in_features = trace_length
    n_layers_s = hp_search_s['n_layers_s']
    n_neurons_s = hp_search_s['n_neurons_s']
    activation_function_s = [x.lower() for x in hp_search_s['activation_function_s']]

    n_layers = trial.suggest_int("n_layers", int(n_layers_s[0]), int(n_layers_s[1]), int(n_layers_s[2]))
    activation_function = trial.suggest_categorical("activation_function", activation_function_s)
    try:
        act = {"relu":nn.ReLU, "tanh":nn.Tanh, "selu":nn.SELU}[activation_function]
    except:
        logger.error('Activation function is not supported now..')
        sys.exit(-1)
    layers = []
    layers.append(nn.Flatten())
    layers.append(nn.BatchNorm1d(trace_length))
    for i in range(n_layers):
        out_features = trial.suggest_int('n_units_l{}'.format(i), int(n_neurons_s[0]), int(n_neurons_s[1]), int(n_neurons_s[2]))
        layers.append(nn.Linear(in_features, out_features))
        layers.append(act())
        layers.append(nn.Dropout(0.1))
        in_features = out_features
    layers.append(nn.Linear(in_features, num_classes))
    return nn.Sequential(*layers)

def make_CNN(trial, hp_search_s, trace_length, num_classes, **kwargs):
    in_features = trace_length
    n_layers_s = hp_search_s['n_layers_s']
    n_neurons_s = hp_search_s['n_neurons_s']
    # n_filters_s = hp_search_s['n_filters_s']
    kernel_size_s = hp_search_s['kernel_size_s']
    activation_function_s = [x.lower() for x in hp_search_s['activation_function_s']]
    n_conv_block_s = hp_search_s['n_conv_block_s']
    #default conv1d parameters
    padding = 0
    stride = 1
    dilation = 1

    n_conv_block = trial.suggest_int('n_conv_block', int(n_conv_block_s[0]), int(n_conv_block_s[1]), int(n_conv_block_s[2]))
    activation_function = trial.suggest_categorical("activation_function", activation_function_s)
    try:
        act = {"relu":nn.ReLU, "tanh":nn.Tanh, "selu":nn.SELU}[activation_function]
    except:
        logger.error('Activation function is not supported now..')
        sys.exit(-1)
    i=0
    layers = []
    layers.append(nn.BatchNorm1d(1))
    for _ in range(n_conv_block):
        kernel_size = trial.suggest_int('kernel_size_{}'.format(i), int(kernel_size_s[0]), int(kernel_size_s[1]), int(kernel_size_s[2]))
        c_in = min(512,8**i)
        c_out = min(512,8**(i+1))
        layers.append(nn.Conv1d(c_in, c_out, kernel_size=kernel_size))
        out_features = (in_features + 2 * padding - dilation * (kernel_size - 1) - 1) // stride + 1
        layers.append(act())
        if i%2==1:
            layers.append(nn.BatchNorm1d(c_out))
        layers.append(nn.AvgPool1d(2))
        out_features = max(1, (out_features - 2) // 2 + 1)
        i+=1
        in_features = out_features
    # layers.append(Dropout(0.5))
    layers.append(nn.Flatten())
    in_features = min(512,8**(i))*(in_features)#Number of filters * number of features
    n_layers = trial.suggest_int("n_layers", int(n_layers_s[0]), int(n_layers_s[1]), int(n_layers_s[2]))
    for i in range(n_layers):
        out_features = trial.suggest_int('n_units_l{}'.format(i), int(n_neurons_s[0]), int(n_neurons_s[1]), int(n_neurons_s[2]))
        layers.append(nn.Linear(in_features, out_features))
        layers.append(act())
        in_features = out_features
    layers.append(nn.Linear(in_features, num_classes))
    return nn.Sequential(*layers)

def make_CNN_simpler(trial, hp_search_s, trace_length, num_classes, **kwargs):
    in_features = trace_length
    n_layers_s = hp_search_s['n_layers_s']
    n_neurons_s = hp_search_s['n_neurons_s']
    activation_function_s = [x.lower() for x in hp_search_s['activation_function_s']]
    #default conv1d parameters
    padding = 0
    stride = 1
    dilation = 1

    activation_function = trial.suggest_categorical("activation_function", activation_function_s)
    try:
        act = {"relu":nn.ReLU, "tanh":nn.Tanh, "selu":nn.SELU}[activation_function]
    except:
        logger.error('Activation function is not supported now..')
        sys.exit(-1)

    i=0
    layers = []
    layers.append(nn.BatchNorm1d(1))
    cin = 1
    for i in range(3):
        kernel_size = 80//(i+1)
        stride = 1
        cout = 128
        layers.append(nn.Conv1d(cin, cout, kernel_size=kernel_size, stride=stride))
        layers.append(act())
        in_features = (in_features + 2 * padding - dilation * (kernel_size - 1) - 1) // stride + 1
        layers.append(nn.BatchNorm1d(cout))
        layers.append(nn.AvgPool1d(2))
        in_features = max(1, (in_features - 2) // 2 + 1)
        cin = cout

    for i in range(2):
        kernel_size = 5
        stride = 1
        cout = 256
        layers.append(nn.Conv1d(cin, cout, kernel_size=kernel_size, stride=stride))
        layers.append(act())
        in_features = (in_features + 2 * padding - dilation * (kernel_size - 1) - 1) // stride + 1
        layers.append(nn.BatchNorm1d(cout))
        layers.append(nn.AvgPool1d(2))
        in_features = max(1, (in_features - 2) // 2 + 1)
        cin = cout

    for _ in range(2):
        kernel_size = 3
        stride = 1
        cout = 512
        layers.append(nn.Conv1d(cin, cout, kernel_size=kernel_size, stride=stride))
        layers.append(act())
        in_features = (in_features + 2 * padding - dilation * (kernel_size - 1) - 1) // stride + 1
        layers.append(nn.BatchNorm1d(cout))
        layers.append(nn.AvgPool1d(2))
        in_features = max(1, (in_features - 2) // 2 + 1)
        cin = cout
    
    layers.append(nn.Flatten())
    in_features = cout*(in_features)#Number of filters * number of features
    n_layers = trial.suggest_int("n_layers", int(n_layers_s[0]), int(n_layers_s[1]), int(n_layers_s[2]))
    for i in range(n_layers):
        out_features = trial.suggest_int('n_units_l{}'.format(i), int(n_neurons_s[0]), int(n_neurons_s[1]), int(n_neurons_s[2]))
        layers.append(nn.Linear(in_features, out_features))
        layers.append(act())
        in_features = out_features
    layers.append(nn.Linear(in_features, num_classes))
    return nn.Sequential(*layers)

def objective_hp(parameters, hp_search_s, trace_data, callbacks=[Callbacks.Callback()], verbose=0, *args, **kwargs):
    hp_search_s = {**hp_search_s_default, **hp_search_s}
    logger.info("\nHyperparameters search space:")
    #TODO: print values without the string representation, maybe use strip("'") for elements in list
    for key in hp_search_s.keys():
        logger.info("{}: {}".format(key,hp_search_s[key]))
    logger.info("\n")

    #Create dataloaders
    trainset = torch.utils.data.TensorDataset(torch.Tensor(trace_data["x_train"]), torch.Tensor(trace_data["m_train"]["label"]))
    valset = torch.utils.data.TensorDataset(torch.Tensor(trace_data["x_val"]), torch.Tensor(trace_data["m_val"]["label"]))
    valloader = torch.utils.data.DataLoader(valset, batch_size=1024, shuffle=True, pin_memory=True, num_workers=4)#no need for gradient descent
    def objective(trial):
        epochs = trial.suggest_int("epochs", hp_search_s['epochs'][0], hp_search_s['epochs'][1], hp_search_s['epochs'][2])
        bs = trial.suggest_int("bs", hp_search_s['bs'][0], hp_search_s['bs'][1], hp_search_s['bs'][2])
        trainloader = torch.utils.data.DataLoader(trainset, batch_size=bs, shuffle=True, pin_memory=True, num_workers=4)

        model = eval(kwargs['model_name'])(trial, hp_search_s, parameters['n_samples'], parameters['num_classes'], **kwargs).to(device)
        
        #Give a new writer to the callbacks for a given trial
        for c in callbacks:
            c.reset(epochs=epochs, trial=trial)
        
        optimizer_name = trial.suggest_categorical("optimizer", hp_search_s['optimizer_s'])
        if hp_search_s['lr_s'][2] == 'log':
            lr = trial.suggest_float("lr", float(hp_search_s['lr_s'][0]), float(hp_search_s['lr_s'][1]), log=True)
        else:
            lr = trial.suggest_float("lr", float(hp_search_s['lr_s'][0]), float(hp_search_s['lr_s'][1]), step=hp_search_s['lr_s'][2])
        optimizer = getattr(optim, optimizer_name)(model.parameters(), lr=lr, betas=(0.9, 0.999), weight_decay=10e-5)

        model.apply(Utils.he_uniform_init_weights)
        history = {'loss':torch.zeros((epochs), device=device), 'accuracy':torch.zeros((epochs), device=device), 'val_loss':torch.zeros((epochs), device=device), 'val_accuracy':torch.zeros((epochs), device=device)}

        loss_function = trial.suggest_categorical("loss_function", hp_search_s['loss_function_s'])
        criterion = {"crossentropy":nn.CrossEntropyLoss(), "mse":nn.MSELoss()}[loss_function.lower()]
        m = nn.LogSoftmax(dim=1)

        model.train()
        skip = [cb.on_train_begin() for cb in callbacks]
        for epoch in range(epochs):
            start_t = time.time()
            skip = [cb.on_epoch_begin(trial=trial) for cb in callbacks]
            for i, (data, labels) in enumerate(trainloader):
                skip = [cb.on_batch_begin() for cb in callbacks]
                inputs, labels = data.to(device), labels.to(device)
                # forward + backward + optimize
                optimizer.zero_grad()
                outputs = model(inputs)
                loss = criterion(m(outputs), labels.long())
                skip = [cb.on_loss_end() for cb in callbacks]
                loss.backward()
                skip = [cb.on_step_begin() for cb in callbacks]
                optimizer.step()
                skip = [cb.on_step_end() for cb in callbacks]
                history['loss'][epoch] += loss.detach().detach()
                history['accuracy'][epoch] += (outputs.argmax(dim=1) == labels).sum()
            history['loss'][epoch] /= (i+1)
            history['accuracy'][epoch] /= parameters['training_size']

            #Validation pass
            with torch.no_grad():
                for i, (data, labels) in enumerate(valloader):
                    inputs, labels = data.to(device), labels.to(device)
                    # forward only
                    outputs = model(inputs)
                    loss = criterion(m(outputs), labels.long())
                    history['val_loss'][epoch] += loss
                    history['val_accuracy'][epoch] += (outputs.argmax(dim=1) == labels).sum()
                history['val_loss'][epoch] /= (i+1)
                history['val_accuracy'][epoch] /= parameters['val_size']
                skip = [cb.on_epoch_end(model=model, epoch=epoch, history=history, trial=trial, last_epoch=epochs, **kwargs) for cb in callbacks]

            # Get the objective value from the metric, and report to trial
            if 'metric' in kwargs.keys():
                if kwargs['metric'] in ['val_accuracy', 'val_loss', 'accuracy', 'loss']:
                    objective_value = history[kwargs['metric']][epoch]
                    trial.report(objective_value, epoch)
                elif kwargs['metric'] == 'gge':
                    # objective value is evaluated inside the callback, and reported here
                    for cb in callbacks:
                        if cb.name == 'gge':
                            objective_value = cb.get_fit_value()
                else:
                    logging.warning("Metric '{}' is not defined, default to objective_value=0".format(kwargs['metric']))
                    objective_value = 0 #No value to report
            
            # Handle pruning based on the intermediate value.
            if trial.should_prune():
                raise optuna.exceptions.TrialPruned()
            
            #deserialize history
            np_history = {}
            for key in history.keys():
                np_history[key] = history[key].detach().cpu().numpy()
            torch.cuda.empty_cache()
            if verbose > 0:
                logger.info("Epoch {}/{} - {:.3f}s - loss: {:.4f} - accuracy: {:.4f} - val_loss: {:.4f} - val_accuracy: {:.4f}".format(epoch+1, epochs, time.time()-start_t, history['loss'][epoch], history['accuracy'][epoch], history['val_loss'][epoch], history['val_accuracy'][epoch]))
        if verbose == 0:
            logger.info("\nEpoch {}/{} - {:.3f}s - loss: {:.4f} - accuracy: {:.4f} - val_loss: {:.4f} - val_accuracy: {:.4f}".format(epoch+1, epochs, time.time()-start_t, history['loss'][epoch], history['accuracy'][epoch], history['val_loss'][epoch], history['val_accuracy'][epoch]))
        skip = [cb.on_train_end(model, trial=trial) for cb in callbacks]
        #Return one float value as final score
        return objective_value
    return objective



# eSCAPyDL - a Deep Learning Side-Channel Analysis Python Framework
# Copyright (C) 2023  Weissbart Léo

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sys, os
import importlib
import logging 
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
import ast
import trsfile
import h5py
import torch
import numpy as np
from tqdm import tqdm 
import Callbacks
from Utils import pbar

from sklearn.preprocessing import normalize

class Dataset:
    """
    Class to construct the traceset and easily get traces, labels and metadata without explicitly calling a specific dataset.
    (enable customizable shuffling + higher level functions for operation on traces)
    """
    def __init__(self, trs_dataset, dl_type, json_data={}):
        self.trs_dataset = trs_dataset
        self.dl_type = dl_type

        if json_data == {}:
            try:
                logger.debug("Loading info on dataset from datasets/{}.py".format(trs_dataset))
                module = importlib.import_module("datasets.{}".format(trs_dataset))
                parameters = getattr(module,"parameters")
                callbacks = getattr(module,"callbacks")
                model_params = getattr(module,"model")
                try:
                    self.hp_search_s = getattr(module,"hp_search_s")
                except AttributeError:
                    self.hp_search_s = {}
                get_attack_parameters = getattr(module,"get_attack_parameters")
                get_known_key = getattr(module,"get_known_key")
                intermediate_value = getattr(module,"intermediate_value")
            except AttributeError as e:
                logger.error(e)
                sys.exit("Dataset is incorrect. Check your command argument.")
        else:
            try:
                logger.debug("Loading info on dataset from loaded json file")
                model_params = json_data["model"]
                parameters = json_data["parameters"]
                callbacks = json_data["callbacks"]
                try:
                    parameters["dataset_name"] = eval(parameters["dataset_name"])
                except Exception:
                    parameters["dataset_name"] = parameters["dataset_name"].strip('"').strip("'")
                parameters["dataset_size"] = int(parameters["dataset_size"])
                parameters["training_size"] = int(parameters["training_size"])
                parameters["val_size"] = int(parameters["val_size"])
                parameters["subkey"] = ast.literal_eval(parameters["subkey"])
                parameters["num_key_hypotheses"] = int(parameters["num_key_hypotheses"])
                parameters["n_samples"] = int(parameters["n_samples"])
                parameters["shuffle"] = bool(parameters["shuffle"])
                self.hp_search_s = json_data["hp_search_s"]
                _locals = {}
                for arg_import in json_data["extra_imports"]:
                    exec(arg_import, globals())
                    exec(arg_import, globals(), _locals)
                for function in json_data["functions"]:
                    exec(json_data["functions"][function], globals())
                    exec(json_data["functions"][function], globals(), _locals)
                get_attack_parameters = _locals["get_attack_parameters"]
                get_known_key = _locals["get_known_key"]
                intermediate_value = _locals["intermediate_value"]
            except Exception as e:
                logger.error(e)
                sys.exit("JSON file is incorrect. Check your json data for missing key argument.")

        self.model_params = model_params
        #Redefine function in utils module
        Callbacks.intermediate_value = intermediate_value
        Callbacks.get_attack_parameters = get_attack_parameters
        Callbacks.get_known_key = get_known_key
        self.parameters = parameters
        self.callbacks = callbacks
    
    def get_parameters(self):
        return self.parameters
    
    def get_model_parameters(self):
        return self.model_params
    
    def get_callbacks(self):
        return self.callbacks
    
    def get_hp_search_s(self):
        return self.hp_search_s

    def traceset(self):
        """
        Class to construct the traceset and easily get traces, labels and metadata without explicitly calling a specific dataset.
        (enable customizable shuffling + higher level functions for operation on traces)
        """
        #TODO: add support for other file formats
        traces, metadata, e_dataset_size = read_dataset(self.parameters)

        PREPROCESS = False
        if PREPROCESS:
            traces = normalize(traces)
            # traces = traces + np.random.normal(0,1,size=traces.shape) # Add random noise
            # traces = np.array([np.sum(traces[:,i:i+5],axis=1) for i in range(0,traces.shape[1],5)]).T # average window resample
            self.parameters["n_samples"] = traces.shape[1]


        logger.info("Traceset loaded")
        #Recompute labels based on intermediate value model
        if isinstance(self.parameters["subkey"], int):
            metadata['label'] = np.array([Callbacks.intermediate_value(Callbacks.get_known_key(metadata, self.parameters["subkey"], i),Callbacks.get_attack_parameters(metadata, self.parameters["subkey"], i)) for i in pbar(range(e_dataset_size), desc="Compute labels")])
        elif isinstance(self.parameters["subkey"], list):
            metadata['label'] = np.array([[Callbacks.intermediate_value(Callbacks.get_known_key(metadata,sk,i),Callbacks.get_attack_parameters(metadata,sk,i)) for sk in self.parameters['subkey']] for i in pbar(range(e_dataset_size), desc="Compute labels")])
        self.parameters["num_classes"] = len(set(metadata["label"].reshape(-1)))
        logger.info("Labels computed")

        traces = np.array(traces)
        traces = np.expand_dims(traces, axis=1)
        logger.info(traces.shape)

        train_size = self.parameters["training_size"]
        val_size = self.parameters["val_size"]
        assert train_size+val_size <= len(traces), "Error: dataset is too small to be splitted"
        #split dataset into train, val and test
        traces_train = np.array(traces[:train_size], dtype=np.float32)
        traces_val = np.array(traces[train_size:train_size+val_size], dtype=np.float32)
        metadata_train = {k:metadata[k][:train_size] for k in metadata.keys()}
        metadata_val = {k:metadata[k][train_size:train_size+val_size] for k in metadata.keys()}
        #if Shuffle is enabled, shuffle the dataset and keep the same order for traces and metadata
        if self.parameters["shuffle"]:
            perm_train = np.random.permutation(train_size)
            perm_val = np.random.permutation(val_size)
            traces_train = traces_train[perm_train]
            traces_val = traces_val[perm_val]
            metadata_train = {k:metadata_train[k][perm_train] for k in metadata_train.keys()}
            metadata_val = {k:metadata_val[k][perm_val] for k in metadata_val.keys()}
        logger.info("Dataset splitted into train and validation sets")

        # logger.debug(metadata_val.keys())
        for i in range(val_size):
            assert (metadata_val["key"][0]).all() == (metadata_val["key"][i]).all(), "Validation set should be on fixed key: {}".format(i)

        #return a dictionary with the traces and metadata
        return  {
            "x_train" : torch.from_numpy(traces_train),
            "x_val" : torch.from_numpy(traces_val),
            "m_train" : {k:torch.from_numpy(metadata_train[k]) for k in metadata_train.keys()},
            "m_val" : {k:torch.from_numpy(metadata_val[k]) for k in metadata_val.keys()},
            }


def read_dataset(parameters):
    if "validation_dataset" in parameters and parameters["validation_dataset"]!=None:
        dataset = Dataset_raw(parameters["dataset_name"], parameters["training_size"])
        assert dataset.len_train_dataset() >= parameters["training_size"]
        val_dataset = Dataset_raw(parameters["validation_dataset"], parameters["val_size"])
        assert val_dataset.len_attack_dataset() >= parameters["val_size"]
        e_dataset_size = parameters["training_size"] + parameters["val_size"]
        traces = np.array([dataset.get_train_trace(i) for i in pbar(range(parameters["training_size"]),desc='Loading training traces')] + [val_dataset.get_attack_trace(i) for i in pbar(range(parameters["val_size"]),desc='Loading validation traces')])
        assert dataset.metadata_keys == val_dataset.metadata_keys, "Parameters in the two datasets must match"
        metadata = {k:np.empty((e_dataset_size, len(dataset.get_train_metadata(0,k))), dtype=np.uint8) for k in dataset.metadata_keys}
        for i in pbar(range(e_dataset_size), desc='Loading metadata'):
            for k in list(dataset.metadata_keys):
                if k<parameters["training_size"]:
                    metadata[k][i] = dataset.get_train_metadata(i,k)
                else:
                    metadata[k][i] = val_dataset.get_attack_metadata(i-parameters["training_size"],k)
    else:
        dataset = Dataset_raw(parameters["dataset_name"], parameters["dataset_size"])
        e_dataset_size = parameters["training_size"] + parameters["val_size"]
        assert int(parameters['dataset_size']) >= e_dataset_size
        traces = np.array([dataset.get_train_trace(i) for i in pbar(range(parameters["training_size"]), desc='Loading training traces')] + [dataset.get_attack_trace(i) for i in pbar(range(parameters["val_size"]),desc='Loading validation traces')])
        metadata = {k:np.empty((e_dataset_size, len(dataset.get_train_metadata(0,k))), dtype=np.uint8) for k in dataset.metadata_keys}
        for i in pbar(range(e_dataset_size), desc='Loading metadata'):
            for k in list(dataset.metadata_keys):
                if i<parameters["training_size"]:
                    metadata[k][i] = dataset.get_train_metadata(i,k)
                else:
                    metadata[k][i] = dataset.get_attack_metadata(i-parameters["training_size"],k)
    return traces, metadata, e_dataset_size
    
class Dataset_raw:
    """Define a generic class to load a dataset from different file formats (trs, h5)
    """
    def __init__(self,filepath:str, dataset_size:int) -> None:
        self.file_type = filepath.split(".")[-1]
        self.supported_file_types = ["trs","h5"]
        self._load_file(filepath)
        self.trace_parameters = self._load_metadata_keys()
        self.dataset_size = dataset_size

    def _load_file(self,filepath):
        if self.file_type == "trs":
            self.file = trsfile.trs_open(filepath,'r')
        elif self.file_type == "h5":
            self.file = h5py.File(filepath,'r')
        else:
            raise NotImplementedError("Only .trs and .h5 files are supported for now")
        
    def _load_metadata_keys(self):
        if self.file_type == "trs":
            self.metadata_keys = list(self.file.get_header(trsfile.Header.TRACE_PARAMETER_DEFINITIONS).keys())
        elif self.file_type == "h5":
            self.metadata_keys = list(self.file['Profiling_traces/metadata'].dtype.names)
        
    def len_attack_dataset(self):
        if self.file_type == "trs":
            return self.file.get_header(trsfile.Header.NUMBER_TRACES)
        elif self.file_type == "h5":
            return len(self.file["Attack_traces/traces"])
        
    def len_train_dataset(self):
        if self.file_type == "trs":
            return self.file.get_header(trsfile.Header.NUMBER_TRACES)
        elif self.file_type == "h5":
            return len(self.file["Profiling_traces/traces"])
        
    def get_train_trace(self, index):
        if self.file_type == "trs":
            return self.file[index].samples
        elif self.file_type == "h5":
            return self.file["Profiling_traces/traces"][index]
        
    def get_attack_trace(self, index):
        if self.file_type == "trs":
            return self.file[self.dataset_size-1 - index].samples
        elif self.file_type == "h5":
            return self.file["Attack_traces/traces"][index]
        
    def get_train_metadata(self, index, key):
        if self.file_type == "trs":
            return self.file[index].parameters[key].value
        elif self.file_type == "h5":
            return self.file["Profiling_traces/metadata"][index][key]
        
    def get_attack_metadata(self, index, key):
        if self.file_type == "trs":
            return self.file[self.dataset_size-1 - index].parameters[key].value
        elif self.file_type == "h5":
            return self.file["Attack_traces/metadata"][index][key]

